# this is for the rare cases where standard plate solving fails
# we use our gaia stars and detections to create a reasonable WCS for our frames.
# to be attempted if all else fails.
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.wcs import WCS
from astropy.table import Table
from astropy.io import fits
import astroalign as aa

from ..structure.database import execute_sqlite_query, get_pandas
from ..utilities.gaia import find_gaia_stars
from ..plotting.sources_plotting import plot_coordinates_and_sources_on_image
from ..processes.plate_solving import post_plate_solve_steps
from ..structure.user_config import get_user_config


def create_initial_wcs(pixel_scale, image_shape, center_ra, center_dec, rotation_angle_deg):
    """
    This makes an astropy wcs object with cd matrix, starting from a pixel scale, image shape, rotation and
    coordinates.

    Args:
        pixel_scale: float, arcsecond / pixel
        image_shape: tuple
        center_ra: float
        center_dec: float
        rotation_angle_deg: float

    Returns:
        astropy.wcs.WCS object
    """
    w = WCS(naxis=2)
    w.wcs.crpix = [(image_shape[1] - 1) / 2, (image_shape[0] - 1) / 2]  # center of the image
    w.wcs.crval = [center_ra, center_dec]  # ra, dec at the center
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]

    rotation_angle_rad = np.deg2rad(rotation_angle_deg)
    pixel_scale_deg = pixel_scale / 3600.0
    w.wcs.cd = np.array([[-pixel_scale_deg * np.cos(rotation_angle_rad), pixel_scale_deg * np.sin(rotation_angle_rad)],
                         [pixel_scale_deg * np.sin(rotation_angle_rad), pixel_scale_deg * np.cos(rotation_angle_rad)]])

    return w


def refine_wcs_with_astroalign(sources, gaia_star_coords, wcs):
    """
    takes in a 'guess' wcs object (made by a helper function in this file), crossmatches the resulting
    pixel positions of the gaia stars to the extracted sources, and if possible calculates a WCS.
    Args:
        sources: astropy Table of sources, with 'xcentroid' and 'ycentroid' columns
        gaia_star_coords: astropy SkyCoord containing the coordinates of our stars (please correct for proper motion)
        wcs: an initial guess wcs object

    Returns:

    """

    gaia_pix_coords = wcs.world_to_pixel(gaia_star_coords)
    gaia_pix_positions = np.vstack(gaia_pix_coords).T

    image_positions = np.column_stack((sources['x'], sources['y']))

    try:
        transformation, (source_idx, gaia_idx) = aa.find_transform(image_positions, gaia_pix_positions)
    except Exception as e:
        print(f"Error finding transformation with astroalign: {e}")
        raise

    crpix = transformation.inverse.params[:2, :2] @ wcs.wcs.crpix + transformation.inverse.translation

    new_cd = transformation.inverse.params[:2, :2] @ wcs.wcs.cd
    wcs_updated = wcs.deepcopy()
    wcs_updated.wcs.crpix = crpix
    wcs_updated.wcs.cd = new_cd

    return wcs_updated


def alternate_plate_solve():
    """
      This cross-matches Gaia detections with the sources in our images, given a good estimation of the pixel scale,
      rotation of the field and center of the field.
      Then, it creates a WCS that best matches the Gaia sources to our pixels.
    Returns:
        Nothing, but loops over the frames and updates database and fits headers.

    """
    user_config = get_user_config()
    ra, dec = user_config['ROI_ra_deg'], user_config['ROI_dec_deg']
    center_radius = {'center': (ra, dec), 'radius':  user_config['alternate_plate_solve_gaia_radius']/3600.}
    gaia_stars = find_gaia_stars('circle', center_radius=center_radius)
    gaia_stars['pmra'][np.isnan(gaia_stars['pmra'])] = 0
    gaia_stars['pmdec'][np.isnan(gaia_stars['pmdec'])] = 0
    gaia_coords = SkyCoord(ra=gaia_stars['ra'],
                           dec=gaia_stars['dec'],
                           pm_ra_cosdec=gaia_stars['pmra'],
                           pm_dec=gaia_stars['pmdec'],
                           frame='icrs',
                           obstime=Time(gaia_stars['ref_epoch'].value, format='decimalyear'))
    pixel_scale = np.mean(user_config['plate_scale_interval'])

    frames_to_process = get_pandas(columns=['id', 'image_relpath', 'sources_relpath', 'mjd'],
                                   conditions=['plate_solved = 0', 'eliminated = 0'])
    plot_dir = user_config['plots_dir'] / 'gaia_plate_solve_diagnostic'
    plot_dir.mkdir(parents=True, exist_ok=True)
    for i, frame in frames_to_process.iterrows():

        gaia_coords_moved = gaia_coords.apply_space_motion(new_obstime=Time(frame['mjd'], format='mjd'))
        frame_path = user_config['workdir'] / frame['image_relpath']
        frame_id = frame['id']
        image = fits.getdata(frame_path).astype(float)
        sources_path = user_config['workdir'] / frame['sources_relpath']
        sources = Table(fits.getdata(sources_path))
        initial_wcs = create_initial_wcs(pixel_scale=pixel_scale, center_ra=ra, center_dec=dec, rotation_angle_deg=0,
                                         image_shape=image.shape)
        try:
            new_wcs = refine_wcs_with_astroalign(sources, gaia_coords_moved, initial_wcs)
            success = True
        except Exception as e:
            print(f"Could not solve frame {frame_id}: {e}.")
            success = False
        if success:
            with fits.open(frame_path, mode="update") as hdul:
                hdul[0].header.update(new_wcs.to_header())
                hdul.flush()
            plot_path = plot_dir / f"{frame_path.stem}.jpg"
            plot_coordinates_and_sources_on_image(image, sources=sources,
                                                  gaia_coords=gaia_coords_moved, wcs=new_wcs, save_path=plot_path)
            post_plate_solve_steps(frame_path=frame_path, user_config=user_config, frame_id=frame_id)

        # at the end, set the image to plate solved in db
        execute_sqlite_query(query="UPDATE frames SET plate_solved = ? WHERE id = ?",
                             params=(1 if success else 0, frame_id), is_select=False)



import numpy as np
from astropy.stats import sigma_clipped_stats

from ..structure.user_config import get_user_config
from ..structure.database import execute_sqlite_query


def get_chi2_bounds(psf_or_fluxes):
    """
    This is a wrapper around the user config. We need chi2 bounds to only query the frames
    with a good enough PSF fit. There are several options to decide on the bounds:
    Params:
        psf_or_fluxes: string, either 'psf' or 'fluxes'.
    Returns:
        tuple (chi2_min, chi2_max)
    """
    user_config = get_user_config()
    assert psf_or_fluxes in ['psf', 'fluxes'], f'get_chi2_bounds: not something I know of: {psf_or_fluxes}'
    conf = user_config[f'{psf_or_fluxes}_fit_exclude_strategy']
    if conf is None:
        return -np.inf, np.inf
    elif type(conf) is dict:
        # then we're either dealing with bounds, or sigma clip
        assert len(conf.keys()) == 1
        key = list(conf.keys())[0]
        conf, val = key, conf[key]
        if conf == 'sigma_clip':
            # here we need to query the chi2. we'll just query them all to have
            # a feeling of what chi2 values we have, can be a systematic in the noisemaps.
            if psf_or_fluxes == 'psf':
                chi2val = execute_sqlite_query("select chi2 from PSFs", is_select=True, use_pandas=True)
            else:
                chi2val = execute_sqlite_query("select chi2 from star_flux_in_frame",
                                               is_select=True, use_pandas=True)
            mean, median, std = sigma_clipped_stats(chi2val['chi2'], sigma=val)
            chi2_min = median - val * std
            chi2_max = median + val * std
            return chi2_min, chi2_max
        elif conf == 'threshold':
            return val
        else:
            raise RuntimeError(f"Unexpected psf_fit_exclude_strategy: {conf}. valid: None, 'sigma_clip' or 'threshold'")



import matplotlib.pyplot as plt
from astropy.visualization import ZScaleInterval, ImageNormalize
from astropy.visualization.stretch import AsinhStretch


def plot_image(image, wcs=None, save_path=None, colorbar=False, **imshow_kwargs):
    """
    Plot the image with detected sources marked (for debugging).

    Parameters:
    image (numpy.ndarray): Image data, 2D array.
    wcs (astropy.wcs.WCS object): the WCS corresponding to the data. default None.
    save_path: do we save the plot somewhere? default None.
    imshow_kwargs: additional keyword arguments to be passed to matplotlib imshow.
    """
    fig = plt.figure(figsize=(11, 11))
    if wcs is not None:
        ax = plt.subplot(projection=wcs)
    else:
        ax = plt.subplot()
    norm = ImageNormalize(image,
                          interval=ZScaleInterval(contrast=0.1),
                          stretch=AsinhStretch()
                          )
    ax.imshow(image, cmap='gray', origin='lower', norm=norm, **imshow_kwargs)
    if colorbar:
        plt.colorbar()

    if save_path is not None:
        plt.tight_layout()
        plt.savefig(save_path, bbox_inches='tight', pad_inches=0.)
        plt.close()
    else:
        return fig, ax



class NoConfigFilePathInEnvironment(Exception):
    def __init__(self):
        message = """You need to define the path to your configuration file, 
e.g. export LIGHTCURVER_CONFIG="/path/to/user_config.yaml"
Then re-run the pipeline.
"""
        super().__init__(message)


class TaskWasNotSuccessful(Exception):
    def __init__(self, message):
        super().__init__(message)



import string
import itertools


def generate_star_names(n):
    """
    just so we have a nicer way (than gaia ID in a given combined footprint / selection of stars) to refer to our stars
    (assign each a letter / a combination of letters).
    Args:
        n: int, number of labels to generate

    Returns:
        list of strings, e.g. ['a', 'b', 'c', 'd', 'e', 'f', ...]
    """
    def all_strings():
        size = 1
        while True:
            for s in itertools.product(string.ascii_lowercase, repeat=size):
                yield "".join(s)
            size += 1

    return [name for _, name in zip(range(n), all_strings())]




import numpy as np
import ephem
from astropy.coordinates import SkyCoord
from astropy.coordinates import Angle
from astropy import units as u


def calculate_airmass(altitude_degrees: np.array) -> np.array:
    """
    Calculate the airmass using Rozenberg's empirical relation.

    Parameters
    ----------
    altitude_degrees : float or ndarray
        The apparent altitude of the object in degrees. Can be a single value or a NumPy array.

    Returns
    -------
    float or ndarray
        The airmass value(s). Returns -1.0 for altitudes below the horizon (altitude_radians < 0),
        -2.0 for altitudes above 90 degrees (altitude_radians > pi/2), and the airmass according
        to Rozenberg's relation for altitudes in between.

    Notes
    -----
    Rozenberg's empirical relation is defined as:
        X = 1 / (sin(ho) + 0.025 * exp(-11 * sin(ho)))
    where ho is the apparent altitude of the object. This formula is applicable down
    to the horizon (where it gives X = 40).
    """
    altitude_radians = np.radians(np.asarray(altitude_degrees))
    with np.errstate(divide='ignore', invalid='ignore'):
        airmass_values = np.where(
            (altitude_radians < 0),
            -1.0,
            np.where(
                (altitude_radians > np.pi / 2),
                -2.0,
                1.0 / (np.sin(altitude_radians) + 0.025 * np.exp(-11.0 * np.sin(altitude_radians)))
            )
        )
    return airmass_values


def ephemeris(mjd: float,
              ra_object: float, dec_object: float,
              telescope_longitude: float, telescope_latitude: float, telescope_elevation: float) -> dict:
    """
    This function calculates and returns the ephemeris for a given object at a specific time and location.

    Parameters:
    mjd (float): Modified Julian Date for the observation.
    ra_object (float): Right ascension of the observed object in degrees.
    dec_object (float): Declination of the observed object in degrees.
    telescope_longitude (float): Longitude of the telescope's location in degrees.
    telescope_latitude (float): Latitude of the telescope's location in degrees.
    telescope_elevation (float): Elevation of the telescope's location in meters.

    Returns:
    dict: A dictionary containing information about the astro conditions, comments, target, moon, and sun.
        - 'astro_conditions' (bool): Indicates if the astro conditions are favorable.
        - 'comments' (str): Additional comments about the astro conditions.
        - 'target_info' (dict): Information about the target object, including altitude, azimuth, and airmass.
        - 'moon_info' (dict): Information about the moon, including distance from the target, illumination, and altitude.
        - 'sun_info' (dict): Information about the sun, including altitude.
    """

    results = {
        'weird_astro_conditions': False,
        'comments': "",
        'target_info': {},
        'moon_info': {},
        'sun_info': {}
    }

    telescope = ephem.Observer()
    # Convert longitude and latitude to strings
    long_str = Angle(telescope_longitude * u.deg).to_string(unit=u.degree, sep=':')
    lat_str = Angle(telescope_latitude * u.deg).to_string(unit=u.degree, sep=':')
    telescope.long = long_str
    telescope.lat = lat_str
    telescope.elevation = telescope_elevation
    telescope.epoch = ephem.J2000

    jd = mjd + 2400000.5
    djd = jd - 2415020.0
    telescope.date = djd

    target = ephem.FixedBody()
    # make them coordinate strings for pyephem:
    coord = SkyCoord(ra=ra_object*u.degree, dec=dec_object*u.degree)
    ra_str = coord.ra.to_string(unit=u.hour, sep=':')
    dec_str = coord.dec.to_string(unit=u.degree, sep=':')
    target._ra = ra_str
    target._dec = dec_str
    target.compute(telescope)

    moon = ephem.Moon()
    moon.compute(telescope)
    moondist = np.degrees(float(ephem.separation(moon, target)))

    sun = ephem.Sun()
    sun.compute(telescope)

    target_alt_deg = np.degrees(target.alt)
    target_az_deg = np.degrees(target.az)

    airmass = calculate_airmass(target_alt_deg)
    if airmass < 1.0 or airmass > 5.0:
        results['weird_astro_conditions'] = True
        results['comments'] += f"Target altitude: {target_alt_deg:.2f} degrees (airmass {airmass:.2f})."

    moon_distance_deg = np.degrees(ephem.separation(moon, target))
    moon_illumination = moon.phase
    moon_alt_deg = np.degrees(moon.alt)

    sun_alt_deg = np.degrees(sun.alt)
    if sun_alt_deg > 0.0:
        results['weird_astro_conditions'] = True
        results['comments'] += f" Sun altitude: {sun_alt_deg:.2f} degrees."

    # Fill target, moon, and sun info
    results['target_info'] = {'altitude_deg': target_alt_deg,
                              'azimuth_deg': target_az_deg,
                              'airmass': airmass,
                              'moon_dist': moondist}
    results['moon_info'] = {'distance_deg': moon_distance_deg,
                            'illumination': moon_illumination,
                            'altitude_deg': moon_alt_deg}
    results['sun_info'] = {'altitude_deg': sun_alt_deg}

    return results


def estimate_seeing(sources_table):
    """
    logic written by Malte Tewes, https://github.com/mtewes in 2010, for the COSMOULINE pipe of COSMOGRAIL.
    it seems to have worked well for 14 years now, so I'm just keeping the main flow of it.

    this function estimates a seeing value (in pixels) based on a table of sources as extracted by sep.
    we need the FWHM column to be present in the table, hence the table of sources has to come from
    the extract_stars function of this repo.

    :param sources_table: astropy.table.Table containing detections
    :return: float, a seeing value (reasonable FWHM) in pixels.
    """

    fwhms = sources_table['FWHM']

    if len(fwhms) > 10:  # loads of stars
        # We want a crude guess at what kind of range we have to look for stars
        # The goal here is to have a "nice-looking"
        # histogram with a well-defined peak somewhere inside the range.
        min_fwhm = 1.5
        med_fwhm = np.median(fwhms)
        if med_fwhm < min_fwhm:
            med_fwhm = min_fwhm

        wide_stars = 3.0 * med_fwhm

        max_fwhm = 30.0
        if wide_stars < max_fwhm:
            max_fwhm = wide_stars

        # At this point the true seeing should be between min_fwhm and max_fwhm.
        # We build a first histogram :
        (hist, edges) = np.histogram(fwhms, bins=10,
                                     range=(min_fwhm, max_fwhm))
        # Note that points outside the range are not taken into account at all,
        # they don't fill the side bins!

        # We find the peak, and build a narrower hist around it
        max_pos = np.argmax(hist)
        if max_pos == 0:
            seeing_pixels = np.median(fwhms)
        elif max_pos == len(hist) - 1:
            seeing_pixels = np.median(fwhms)
        else:  # the normal situation :
            peak_pos = float(0.5 * (edges[max_pos] + edges[max_pos + 1]))

            # We build a second histogram around this position,
            # with a narrower range:
            hist, edges = np.histogram(fwhms,
                                       bins=10,
                                       range=(peak_pos - 2.0, peak_pos + 2.0))
            max_pos = np.argmax(hist)
            peak_pos = 0.5 * (edges[max_pos] + edges[max_pos + 1])

            # We take the median of values around this peak_pos :
            star_fwhms = fwhms[np.logical_and(fwhms > peak_pos - 1.0,
                                              fwhms < peak_pos + 1.0)]
            if len(star_fwhms) > 0:
                seeing_pixels = np.median(star_fwhms)
            else:
                seeing_pixels = peak_pos

    elif len(fwhms) > 0:  # few stars, not ideal but whatever
        seeing_pixels = np.median(fwhms)

    else:  # no stars
        seeing_pixels = -1.0
    return seeing_pixels



import numpy as np
from datetime import datetime
from pathlib import Path
from astropy.io import fits
import h5py
import json
from copy import deepcopy
import pandas as pd
from scipy.ndimage import shift, rotate
from astropy.wcs import WCS
from astropy.coordinates import SkyCoord
from astropy import units as u
from photutils.aperture import CircularAperture, aperture_photometry
import logging

from starred.deconvolution.deconvolution import setup_model
from starred.deconvolution.loss import Loss
from starred.optim.optimization import Optimizer
from starred.deconvolution.parameters import ParametersDeconv
from starred.utils.noise_utils import propagate_noise

from ..structure.user_config import get_user_config
from ..structure.database import get_pandas, execute_sqlite_query
from ..utilities.footprint import get_combined_footprint_hash
from ..utilities.starred_utilities import get_flux_uncertainties
from ..plotting.star_photometry_plotting import plot_joint_deconv_diagnostic


def align_data_interpolation(array, starred_kwargs):
    """
    This is a utility function that takes in the data we use for deconvolution, and
    - de-rotates it
    - 'de-translates' it -- both according to our fitted values of translation and rotation.
    It uses interoplation to do so, hence to be used as a diagnostic tool only.

    Parameters:
        array: array containing our data, typically shape (N_epochs, nx, ny)
        starred_kwargs: the STARRED optimized keywords, that contain the rotations and translations.


    Returns:
        array_shift: the same array, but derotated and de-translated
    """
    dx = starred_kwargs['kwargs_analytic']['dx']
    dy = starred_kwargs['kwargs_analytic']['dy']
    alpha = starred_kwargs['kwargs_analytic']['alpha']

    array_shift = np.array(
        [rotate(shift(a, (-ddy, -ddx)), alph, reshape=False) for a, ddx, ddy, alph in zip(array, dx, dy, alpha)]
    )

    return array_shift


def do_deconvolution_of_roi():
    """
    Optionally called by the workflow manager.
    This is probably a bit too rigid given how complicated the joint modelling of 1000+ epochs of a blended ROI
    can be. But it can be used as a template for your own.
    This simply jointly models your ROI cutouts with the point sources given in the config file.
    Should work for most cases still!
    Returns:
        Nothing
    """
    logger = logging.getLogger('lightcurver.roi_modelling')
    user_config = get_user_config()
    if not user_config['do_ROI_model']:
        # we do nothing, this is an optional step.
        return

    frames_ini = get_pandas(columns=['id'],
                            conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_ini['id'].to_list())
    deconv_file = user_config['prepared_roi_cutouts_path']

    if deconv_file is None:
        roi = user_config['roi_name']
        deconv_file = user_config['workdir'] / 'prepared_roi_cutouts' / f"cutouts_{combined_footprint_hash}_{roi}.h5"

    # load the data
    with h5py.File(deconv_file, 'r') as f:
        data = np.array(f['data'])
        noisemap = np.array(f['noisemap'])
        s = np.array(f['psf'])
        scale = np.nanmax(data)
        data /= scale
        noisemap /= scale
        seeings = np.array(f['seeing'])
        mjds = np.array(f['mjd'])
        zeropoint = np.array(f['global_zeropoint'])
        norm_errs = np.array(f['relative_normalization_error'])
        frame_ids = np.array(f['frame_id'])
        subsampling_factor = np.array(f['subsampling_factor'])
        pixel_scales = np.array(f['pixel_scale'])
        angles_to_north = np.array(f['angle_to_north'])
        wcs = np.array(f['wcs'])

    message = "The PSF models seem to have different subsampling factors! Incompatible with a STARRED deconvolution."
    unique_subsampling = (np.unique(subsampling_factor).size == 1)
    if not unique_subsampling:
        logger.error(message + ' Stopping the pipeline.')
    assert np.unique(subsampling_factor).size == 1, message
    subsampling_factor = int(subsampling_factor[0])
    im_size = data.shape[1]
    im_size_up = subsampling_factor * im_size
    epochs = data.shape[0]

    ps_coords = user_config['point_sources']
    ordered_ps = sorted(ps_coords.keys())
    logger.info(f'Jointly deconvolving {epochs} cutouts from your ROI, including {len(ordered_ps)} point sources.')

    # we use the first WCS as reference.
    # so the starred rotation angles will be w.r.t. to this one as well
    angles_to_north -= angles_to_north[0]
    wcs_ref = WCS(wcs[0].decode('utf-8'))
    ps_pixels = {lab: np.array(wcs_ref.world_to_pixel(SkyCoord(*val, unit=u.degree))).flatten() for lab, val in
                 ps_coords.items()}

    # so, initial guess for the positions of the point sources in pixels:
    xs = np.array([ps_pixels[ps][0] for ps in ordered_ps])
    ys = np.array([ps_pixels[ps][1] for ps in ordered_ps])

    # now, initial guess for the fluxes
    pixel_scale = np.nanmedian(pixel_scales)
    stack = np.nanmedian(data, axis=0)
    radii = 0.66 * np.average(seeings) / pixel_scale  # a bit bigger than FWHM.
    positions = [(x, y) for x, y in zip(xs, ys)]
    apertures = CircularAperture(positions, r=radii)
    photometry = aperture_photometry(stack, apertures)
    aperture_fluxes = list(photometry['aperture_sum'])

    # ok, ready to define the STARRED model
    offset = (im_size - 1) / 2.  # removing half image size, as starred has (0,0) = center of image
    initial_c_x = xs - offset
    initial_c_y = ys - offset
    initial_a = list(aperture_fluxes)
    initial_a = len(data) * initial_a
    model, kwargs_init, kwargs_up, kwargs_down, kwargs_fixed = setup_model(data,
                                                                           noisemap ** 2,
                                                                           s,
                                                                           initial_c_x,
                                                                           initial_c_y,
                                                                           subsampling_factor,
                                                                           initial_a)
    # prepare for the random rotations of the pointing ...
    kwargs_init['kwargs_analytic']['alpha'] = angles_to_north
    kwargs_fixed['kwargs_analytic']['alpha'] = angles_to_north
    # if we provide a background:
    if user_config['starting_background'] is not None:
        bck_path = Path(user_config['starting_background'])
        if not bck_path.is_absolute():
            bck_path = user_config['work_dir'] / bck_path
        if bck_path.name.endswith('fits'):
            bck = fits.getdata(bck_path)
        else:
            bck = np.load(bck_path)
        h = bck.flatten() / scale
        kwargs_init['kwargs_background']['h'] = h
        kwargs_fixed['kwargs_background']['h'] = h

    # ok, first step: translations and fluxes
    kwargs_fixed = deepcopy(kwargs_init)
    del kwargs_fixed['kwargs_analytic']['dx']
    del kwargs_fixed['kwargs_analytic']['dy']
    del kwargs_fixed['kwargs_analytic']['a']
    parameters = ParametersDeconv(kwargs_init=kwargs_init,
                                  kwargs_fixed=kwargs_fixed,
                                  kwargs_up=kwargs_up,
                                  kwargs_down=kwargs_down)

    loss = Loss(data, model, parameters, noisemap**2)

    optim = Optimizer(loss, parameters, method='l-bfgs-b')

    best_fit, logL_best_fit, extra_fields, runtime = optim.minimize(maxiter=user_config['roi_deconv_translations_iters'])
    kwargs_partial1 = deepcopy(parameters.best_fit_values(as_kwargs=True))
    logger.info('Finished first optimization, only varying the fluxes of the point sources and the translations.')

    # next, include the background.
    kwargs_fixed = deepcopy(kwargs_partial1)
    if user_config['further_optimize_background']:
        del kwargs_fixed['kwargs_background']['h']
    del kwargs_fixed['kwargs_background']['mean']
    del kwargs_fixed['kwargs_analytic']['a']
    del kwargs_fixed['kwargs_analytic']['c_x']
    del kwargs_fixed['kwargs_analytic']['c_y']
    del kwargs_fixed['kwargs_analytic']['dx']
    del kwargs_fixed['kwargs_analytic']['dy']
    W = propagate_noise(model, noisemap, kwargs_init, wavelet_type_list=['starlet'],
                        method='SLIT', num_samples=500, seed=1, likelihood_type='chi2',
                        verbose=False, upsampling_factor=subsampling_factor)[0]

    parameters = ParametersDeconv(kwargs_init=kwargs_partial1,
                                  kwargs_fixed=kwargs_fixed,
                                  kwargs_up=kwargs_up,
                                  kwargs_down=kwargs_down)

    loss = Loss(data, model, parameters, noisemap**2,
                regularization_terms='l1_starlet',
                regularization_strength_scales=1,
                regularization_strength_hf=1.,
                regularization_strength_positivity=100.,
                W=W)

    optim = Optimizer(loss, parameters, method='adabelief')
    optimiser_optax_option = {
        'max_iterations': user_config['roi_deconv_all_iters'],
        'init_learning_rate': 1e-4, 'schedule_learning_rate': False,
        'restart_from_init': False, 'stop_at_loss_increase': False,
        'progress_bar': True, 'return_param_history': True
    }

    best_fit, logL_best_fit, extra_fields, runtime = optim.minimize(**optimiser_optax_option)
    kwargs_final = deepcopy(parameters.best_fit_values(as_kwargs=True))

    out_dir = deconv_file.parent
    # the easy stuff, let's output the astrometry first:
    x_pixels = np.array(kwargs_final['kwargs_analytic']['c_x'] + kwargs_final['kwargs_analytic']['dx'][0])
    y_pixels = np.array(kwargs_final['kwargs_analytic']['c_y'] + kwargs_final['kwargs_analytic']['dy'][0])
    ps_coords_post = wcs_ref.pixel_to_world_values(np.array((x_pixels, y_pixels)).T)
    ps_coords_post = {ps: list(coord) for ps, coord in zip(ordered_ps, ps_coords_post)}
    with open(out_dir / f'{combined_footprint_hash}_astrometry.json', 'w') as ff:
        json.dump(ps_coords_post, ff)

    # the fluxes ...
    fluxes = kwargs_final['kwargs_analytic']['a']
    # get the uncertainties on the fluxes
    flux_photon_uncertainties = get_flux_uncertainties(kwargs=kwargs_final, kwargs_down=kwargs_down,
                                                       kwargs_up=kwargs_up,
                                                       data=data, noisemap=noisemap, model=model)
    curves = {}
    d_curves = {}
    # let's separate the fluxes by point source
    for i, ps in enumerate(ordered_ps):
        curve = fluxes[i::len(ordered_ps)] * scale
        curve_photon_uncertainties = flux_photon_uncertainties[i::len(ordered_ps)] * scale
        # these need be compounded with the normalisation errors.
        curve_norm_err = norm_errs * curve
        curves[ps] = curve
        d_curves[ps] = (curve_photon_uncertainties**2 + curve_norm_err**2)**0.5

    # ok, onto the chi2
    modelled_pixels = model.model(kwargs_final)
    residuals = data - modelled_pixels
    chi2_per_frame = np.nansum((residuals**2 / noisemap**2), axis=(1, 2)) / model.image_size ** 2

    # let's fold in some info, and save!
    df = []
    num_epochs = len(frame_ids)
    for epoch in range(num_epochs):
        row = {
            'frame_id': frame_ids[epoch],
            'mjd': mjds[epoch],
            'zeropoint': zeropoint,
            'reduced_chi2': chi2_per_frame[epoch],
        }
        for ps in ordered_ps:
            row[f'{ps}_flux'] = curves[ps][epoch]
            row[f'{ps}_d_flux'] = d_curves[ps][epoch]
        df.append(row)
    df = pd.DataFrame(df).set_index('frame_id')
    df.to_csv(out_dir / f'{combined_footprint_hash}_fluxes.csv')
    # ok, now some diagnostics.
    # first, subtract the point sources from the data, see what the stack looks like.
    # (helps to spot PSF model problems, or fit really gone wrong)
    kwargs_only_ps = deepcopy(kwargs_final)
    kwargs_only_ps['kwargs_background']['h'] *= 0.0

    data_no_ps = data - model.model(kwargs_only_ps)
    data_no_ps = align_data_interpolation(data_no_ps, kwargs_only_ps)
    stack_no_ps = np.nanmean(data_no_ps, axis=0)
    fits.writeto(out_dir / f'{combined_footprint_hash}_stack_without_point_sources.fits', scale * stack_no_ps,
                 overwrite=True)

    # and of course, output the deconvolution
    deconv, h = model.getDeconvolved(kwargs_final, 0)
    fits.writeto(out_dir / f'{combined_footprint_hash}_deconvolution.fits', scale * np.array(deconv),
                 overwrite=True)
    fits.writeto(out_dir / f'{combined_footprint_hash}_background.fits', scale * np.array(h),
                 overwrite=True)

    # now a diagnostic plot
    plot_deconv_dir = user_config['plots_dir'] / 'deconvolutions' / str(combined_footprint_hash)
    plot_deconv_dir.mkdir(exist_ok=True, parents=True)
    loss_history = optim.loss_history
    time_now = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
    plot_file = plot_deconv_dir / f"{time_now}_joint_deconv_roi_{roi}.jpg"
    plot_joint_deconv_diagnostic(datas=data, noisemaps=noisemap,
                                 residuals=residuals,
                                 chi2_per_frame=chi2_per_frame, loss_curve=loss_history,
                                 save_path=plot_file)
    logger.info(f'Finished modelling the ROI. Diagnostic plot at {plot_file}. '
                f"The global reduced chi2 was {np.mean(chi2_per_frame):.02f}. ")




import sqlite3
import pandas as pd

from ..structure.user_config import get_user_config


def get_pandas(conditions=None, columns=None, table='frames'):
    """
    Retrieve data from the database based on specified conditions and return as a Pandas DataFrame.

    Parameters:
        conditions (list of str): The SQL conditions to filter the data, default None.
        columns (str or list of str, optional): Columns to select from the database. Defaults to None,
            which selects all columns.
        table: string, which table to fetch data from.
    Returns:
        pandas.DataFrame: A DataFrame containing the results of the SQL query.
    """
    db_path = get_user_config()['database_path']
    conn = sqlite3.connect(db_path)
    if columns is None:
        columns = '*'
    request = f"SELECT {','.join(columns)} FROM {table}"
    if conditions is not None:
        conditions = ' AND '.join(conditions)
        request += f" WHERE {conditions}"

    try:
        df = pd.read_sql_query(request, conn)
        return df
    finally:
        conn.close()


def execute_sqlite_query(query, params=(), is_select=True, timeout=15.0, use_pandas=False):
    """
    The most used interface to our database, very simple:
    connects to the database, and depending of whether we are selecting or inserting returns data or a row count.
    Also, if selecting, option to get back a pandas dataframe by reading the query with pandas.
    Args:
        query: string, the query
        params: tuple, the parameters corresponding to the '?' of the query.
        is_select: bool, are we selecting or inserting?
        timeout: in seconds, keep it large in case of slow storage
        use_pandas: bool, if select, get a pandas table of the data? (else list of tuples)

    Returns:
        either int if is_select=False, or a pandas dataframe if is_select=True and use_pandas=True, or some list of
        tuples.
    """
    db_path = get_user_config()['database_path']
    with sqlite3.connect(db_path, timeout=timeout) as conn:
        cursor = conn.cursor()
        if is_select:
            if use_pandas:
                return pd.read_sql_query(sql=query, con=conn, params=params)
            cursor.execute(query, params)
            return cursor.fetchall()
        else:
            cursor.execute(query, params)
            conn.commit()
            return cursor.rowcount


def get_count_based_on_conditions(conditions, table='frames'):
    """

    Args:
        conditions: string, e.g. 'plate_solved = 1 and eliminated = 0'
        table:

    Returns:

    """
    db_path = get_user_config()['database_path']
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    request = f"select count(*) from {table} where {conditions}"

    return cursor.execute(request).fetchone()[0]


def select_stars(combined_footprint_hash, stars_to_use=None):
    """
    Selects all the stars, either
     -- top 10 closest stars to ROI if stars_to_use is None
     -- top 'stars_to_use' closest stars to ROI if star s_to_use is int
     -- stars whose name is in 'stars_to_use' if stars_to_use is a list.

    Useful for selecting the stars we want to use when calculating a normalization coefficient.

    Args:
        combined_footprint_hash: since the stars were queried in a footprint, specify which. (or might get duplicates or stars
                        that are not in the current footprint)
        stars_to_use:   None or int or list, see docstring

    Returns:
        a pandas dataframe containing our stars (name, id coordinates ...)
    """
    base_query = "SELECT * FROM stars s WHERE combined_footprint_hash = ?"
    if stars_to_use is None:
        stars_to_use = 10  # make it an int for top 10 selection

    if type(stars_to_use) is int:
        # Query for top closest stars
        query = base_query + """
        ORDER BY s.distance_to_roi_arcsec ASC
        LIMIT ?
        """
        params = (combined_footprint_hash, stars_to_use)
    elif type(stars_to_use) is list:
        # Query for stars in the user-defined list
        placeholders = ','.join(['?'] * len(stars_to_use))
        query = base_query + f"""
        AND s.name IN ({placeholders})
        """
        params = (combined_footprint_hash, *stars_to_use)
    else:
        raise RuntimeError(f'stars_to_use argument: expected types None, int or list, got: {type(stars_to_use)}')

    return execute_sqlite_query(query, params, use_pandas=True)


def select_stars_for_a_frame(frame_id, combined_footprint_hash, stars_to_use=None):
    """
    Selects all the stars available in a given frame, either
     -- top 10 closest stars to ROI if stars_to_use is None
     -- top 'stars_to_use' closest stars to ROI if stars_to_use is int
     -- stars whose name is in 'stars_to_use' if stars_to_use is a list.

    Useful for selecting the stars we want to use when modelling the PSF
    or calculating a normalization coefficient.
    
    #TODO very similar to query_stars_for_frame_and_footprint,
    #TODO consider merging in the future.
    Args:
        frame_id:  database frame ID
        combined_footprint_hash: hash of the footprint in which the stars were originally queried.
        stars_to_use:   None or int or list, see docstring

    Returns:
        a pandas dataframe containing our stars (name, id coordinates ...)
    """
    base_query = """
        SELECT 
            sif.frame_id, 
            s.gaia_id, 
            s.name, 
            s.ra, 
            s.dec, 
            s.distance_to_roi_arcsec
        FROM stars_in_frames sif
        JOIN stars s ON sif.star_gaia_id = s.gaia_id AND sif.combined_footprint_hash = s.combined_footprint_hash
        WHERE sif.frame_id = ? AND s.combined_footprint_hash = ?"""
    if stars_to_use is None:
        stars_to_use = 10  # make it an int for top 10 selection

    if type(stars_to_use) is int:
        # Query for top closest stars
        query = base_query + """
        ORDER BY s.distance_to_roi_arcsec ASC
        LIMIT ?
        """
        params = (frame_id, combined_footprint_hash, stars_to_use)
    elif type(stars_to_use) is list:
        # Query for stars in the user-defined list
        placeholders = ','.join(['?'] * len(stars_to_use))
        query = base_query + f"""
        AND s.name IN ({placeholders})
        """
        params = (frame_id, combined_footprint_hash, *stars_to_use)
    else:
        raise RuntimeError(f'stars_to_use argument: expected types None, int or list, got: {type(stars_to_use)}')

    return execute_sqlite_query(query, params, use_pandas=True)


def query_stars_for_frame_and_footprint(frame_id, combined_footprint_hash=None):
    """
    Queries and returns all stars linked to a specific frame, optionally filtered by a specific footprint hash.

    Parameters:
    - db_path: The path to the SQLite database file.
    - frame_id: The ID of the frame for which to find linked stars.
    - combined_footprint_hash: Optional. The combined footprint hash to filter the stars by.

    Returns:
    a pandas dataframe with the stars associated to the frame and footprint
    """

    sql_query = """
    SELECT stars.*
    FROM stars
    INNER JOIN stars_in_frames ON stars.gaia_id = stars_in_frames.star_gaia_id 
                                  AND stars.combined_footprint_hash = stars_in_frames.combined_footprint_hash
    WHERE stars_in_frames.frame_id = ?
    """

    # params for the query
    params = [frame_id]

    # if combined_footprint_hash, add it to the query and params
    if combined_footprint_hash is not None:
        sql_query += "AND stars.combined_footprint_hash = ?"
        params.append(combined_footprint_hash)

    stars = execute_sqlite_query(sql_query, params=params, is_select=True, use_pandas=True)

    return stars


def initialize_database(db_path=None):
    """
    initializes the database we'll be working with to keep track of our frames.

    Parameters:
        :param db_path: string or path, default None. Not used during execution of the pipeline,
                        but can be provided for tests.
    """
    if db_path is None:
        db_path = get_user_config()['database_path']
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    # table of frames
    column_definitions = [
        "id INTEGER PRIMARY KEY",
        "filter TEXT",
        "mjd REAL",
        "exptime REAL",
        "gain REAL",
        "original_image_path TEXT",  # e.g. /some/drive/2023-02-01T01:23:35.fits
        "image_relpath TEXT UNIQUE",  # e.g. frames/2023-02-01T01:23:35.fits -- relative to $workdir
        "sources_relpath TEXT",  # convention: frames/2023-02-01T01:23:35_sources.fits  -- same dir as above
        "telescope_latitude REAL",
        "telescope_longitude REAL",
        "telescope_elevation REAL",
        "telescope_name TEXT",
        "telescope_imager_name TEXT",
        "plate_solved INTEGER DEFAULT 0",
        "attempted_plate_solve INTEGER DEFAULT 0",  # will be set to 1 if we ever attempt a plate solve.
        "pixel_scale REAL DEFAULT NULL",
        "eliminated INTEGER DEFAULT 0",
        "airmass REAL DEFAULT NULL",
        "degrees_to_moon REAL DEFAULT NULL",
        "moon_phase REAL DEFAULT NULL",
        "sun_altitude REAL DEFAULT NULL",
        "seeing_pixels REAL DEFAULT NULL",
        "seeing_arcseconds REAL DEFAULT NULL",
        "sky_level_electron_per_second REAL DEFAULT NULL",
        "background_rms_electron_per_second REAL DEFAULT NULL",
        "ellipticity REAL DEFAULT NULL",
        "azimuth REAL DEFAULT NULL",
        "altitude REAL DEFAULT NULL",
        "comment TEXT DEFAULT NULL",
        "roi_in_footprint INTEGER DEFAULT 0",
        "angle_to_north REAL DEFAULT 0.0",
        # potential new stuff to keep track on in future here
    ]

    # create the 'frames' table if it doesn't exist
    cursor.execute(f"CREATE TABLE IF NOT EXISTS frames ({', '.join(column_definitions)})")

    # add new columns to the 'frames' table if they're not already present
    for column_definition in column_definitions:
        try:
            cursor.execute(f"ALTER TABLE frames ADD COLUMN {column_definition}")
        except sqlite3.OperationalError:
            # operational error if column exists, ignore
            pass

    # now we'll need stars to calibrate the PSF and normalization of each frame.
    # these stars will be queried in a "footprint",  which can be changed to change the selection
    # of reference stars. So we'll refer to the footprint at hand in all steps.
    # a "footprint" is a polygon, composed of vertices (ra,dec).
    # we assume that the gnonomic projection is fine for checking what lands in a footprint
    # first, we define a footprint for each frame
    cursor.execute("""CREATE TABLE IF NOT EXISTS footprints (
                      frame_id INTEGER PRIMARY KEY,
                      polygon TEXT NOT NULL,
                      FOREIGN KEY (frame_id) REFERENCES frames (id)
                      )""")

    # then, a "combined" (intersection or union) footprint, to which we give a hash value that identifies it.
    # (this way, we don't recompute everything if we define a new footprint that actually is identical to one
    # we already processed).
    # this combined footprint is what we will refer to in every downstream step.
    cursor.execute("""CREATE TABLE IF NOT EXISTS combined_footprint ( 
                      id INTEGER PRIMARY KEY, 
                      hash INTEGER UNIQUE,  -- will be a hash of a concatenation of the used frames' ids
                      largest TEXT,
                      common TEXT
                      )""")

    # now the stars:
    # the stars will be filled in once by a python process.
    cursor.execute("""CREATE TABLE IF NOT EXISTS stars (
                      combined_footprint_hash INTEGER, -- which footprint was this star was queried from
                      name TEXT DEFAULT NULL, -- typically a letter
                      ra REAL,
                      dec REAL,
                      gmag REAL,
                      rmag REAL,
                      bmag REAL,               
                      pmra REAL,
                      pmdec REAL,    
                      ref_epoch REAL,   
                      gaia_id TEXT,
                      distance_to_roi_arcsec REAL,
                      FOREIGN KEY (combined_footprint_hash) REFERENCES combined_footprint(hash),
                      PRIMARY KEY (combined_footprint_hash, gaia_id)
                      )""")

    # linking stars and frame
    # again, a python process will check the footprint of each image
    # and fill in this table once for each image, the idea being able to query which stars are available in which image.
    cursor.execute("""CREATE TABLE IF NOT EXISTS stars_in_frames (
                      frame_id INTEGER,
                      star_gaia_id INTEGER,
                      combined_footprint_hash INTEGER,
                      FOREIGN KEY (frame_id) REFERENCES frames(id),
                      FOREIGN KEY (star_gaia_id) REFERENCES stars(gaia_id),
                      FOREIGN KEY (combined_footprint_hash) REFERENCES combined_footprint(hash),
                      PRIMARY KEY (combined_footprint_hash, frame_id, star_gaia_id)
                      )""")

    # here we keep track of which frames have a given PSF.
    # we trace back to which footprint the stars were queried in.
    # the subsampling factor is also defined in yaml file.
    cursor.execute("""CREATE TABLE IF NOT EXISTS PSFs (
                      combined_footprint_hash INTEGER,
                      frame_id INTEGER,
                      chi2 REAL, -- chi2 of the fit of the PSF
                      psf_ref TEXT, -- convention: sorted concatenation of all star names used in the model.
                      subsampling_factor INTEGER,  -- we do starred (pixelated) PSFs.
                      relative_loss_differential REAL, -- absolute change in last 10% of the iterations vs beginning
                      FOREIGN KEY (frame_id) REFERENCES frames(id),
                      FOREIGN KEY (combined_footprint_hash) REFERENCES combined_footprint(hash),
                      PRIMARY KEY (combined_footprint_hash, frame_id, psf_ref)
                      )""")

    # once we'll have PSF models, we'll do PSF photometry of the stars in the field.
    # the table below will keep track of the fitted fluxes.
    cursor.execute("""CREATE TABLE IF NOT EXISTS star_flux_in_frame (
                      frame_id INTEGER,
                      star_gaia_id INTEGER, 
                      combined_footprint_hash INTEGER,
                      flux REAL, -- in e- / second
                      flux_uncertainty REAL,
                      chi2 REAL, -- chi2 of fit in this specific frame.
                      relative_loss_differential REAL, -- absolute change in last 10% of the iterations vs beginning
                      FOREIGN KEY (frame_id) REFERENCES frames(id),
                      FOREIGN KEY (star_gaia_id) REFERENCES stars(gaia_id),
                      FOREIGN KEY (combined_footprint_hash) REFERENCES combined_footprint(hash),
                      PRIMARY KEY (combined_footprint_hash, frame_id, star_gaia_id)
                      )""")

    # very similarly to PSFs, we keep track of normalization coefficients.
    # these are computed once per image within python, from the star fluxes above.
    cursor.execute("""CREATE TABLE IF NOT EXISTS normalization_coefficients (
                      frame_id INTEGER,
                      combined_footprint_hash INTEGER,
                      coefficient REAL,
                      coefficient_uncertainty REAL,
                      FOREIGN KEY (frame_id) REFERENCES frames(id),
                      FOREIGN KEY (combined_footprint_hash) REFERENCES combined_footprint(hash),
                      PRIMARY KEY (combined_footprint_hash, frame_id)
                      )""")

    # zero points are derived from normalization coefficients and are for APPROXIMATE magnitude calibration
    # as we are going to derive them from gaia colors and band conversions.
    cursor.execute("""CREATE TABLE IF NOT EXISTS approximate_zeropoints (
                      frame_id INTEGER,
                      combined_footprint_hash INTEGER,
                      zeropoint REAL,
                      zeropoint_uncertainty REAL,
                      FOREIGN KEY (frame_id) REFERENCES frames(id),
                      FOREIGN KEY (combined_footprint_hash) REFERENCES combined_footprint(hash),
                      PRIMARY KEY (combined_footprint_hash, frame_id)
                      )""")

    conn.commit()
    conn.close()



# here we'll define simple functions that check the health of our products.
# can be beneficial to stop the pipeline if an early stage fails.

from ..structure.database import get_count_based_on_conditions
from ..structure.user_config import get_user_config


def check_plate_solving():
    user_config = get_user_config()
    plate_solved = get_count_based_on_conditions(conditions='plate_solved = 1 and eliminated = 0', table='frames')
    total = get_count_based_on_conditions(conditions='eliminated = 0', table='frames')
    min_success_fraction = user_config['plate_solving_min_success_fraction']
    reasonable_loss = plate_solved / total >= min_success_fraction
    if not reasonable_loss:
        message = "The plate solving failed too often given your config's plate_solving_min_success_fraction"
        message += "There might be a problem, or there might just be a lot of difficult images"
        message += "Please investigate"
    else:
        message = "Plate solving succeeded often enough."
    return reasonable_loss, message





import logging
from astropy.coordinates import SkyCoord
import json
import numpy as np

from ..utilities.footprint import load_combined_footprint_from_db, get_frames_hash
from ..structure.user_config import get_user_config
from ..processes.frame_star_assignment import populate_stars_in_frames
from ..utilities.gaia import find_gaia_stars
from ..utilities.star_naming import generate_star_names
from ..structure.database import get_pandas, execute_sqlite_query
from ..plotting.sources_plotting import plot_footprints_with_stars


def query_gaia_stars():
    """
    this is called by the workflow manager.
    Queries the frames to be used, and checks the star selection strategy from the user config.
    Then queries the stars with the additional criteria from the config (mag range, astrometric noise, photometric
    noise ...) and inserts them in the database for easy access.
    Also, runs the function that assigns stars to each frame depending on the respective footprint of the frame.
    Returns:

    """
    logger = logging.getLogger("lightcurver.querying_ref_stars_from_gaia")
    user_config = get_user_config()
    frames_info = get_pandas(columns=['id', 'pixel_scale'], conditions=['frames.eliminated != 1'])
    if user_config['star_selection_strategy'] != 'ROI_disk':
        # then it depends on the frames we're considering.
        frames_hash = get_frames_hash(frames_info['id'].to_list())
    else:
        # if ROI_disk, it does not depend on the frames: unique region defined by its radius.
        frames_hash = hash(user_config['ROI_disk_radius_arcseconds'])

    # before doing anything, check whether we are already done
    count = execute_sqlite_query("SELECT COUNT(*) FROM stars WHERE combined_footprint_hash = ?",
                                 params=(frames_hash,), is_select=True)[0][0]
    if count > 0 and not user_config['gaia_query_redo']:
        logger.info(f'Gaia stars already fetched for this footprint: {frames_hash}')
        logger.info('Still re-calculating which star is in which frame again.')
        # we still need to populate the new frames though
        populate_stars_in_frames()
        return
    elif count > 0 and user_config['gaia_query_redo']:
        logger.info(f'Gaia stars already fetched for this footprint: {frames_hash} but redo is True.')
        # then we need to purge the database from the stars queried with this footprint.
        # TODO I forgot we have two types of footprints for a given footprint hash dayum
        # TODO for now proceeding with the user having to set redo if changing footprint type
        execute_sqlite_query("DELETE FROM stars WHERE combined_footprint_hash = ?",
                             params=(frames_hash,), is_select=True)
        logger.info(f'  deleted previously queried stars.')

    if user_config['star_selection_strategy'] == 'common_footprint_stars':
        logger.info(f'config star selection strategy: common footprint of the frames')
        _, common_footprint = load_combined_footprint_from_db(frames_hash)
        region_type = 'polygon'
        query_footprint = common_footprint['coordinates'][0]
        # then we want to make sure we use stars that are available in all frames.
        # this likely achieves the best precision, but is only possible typically in dedicated
        # monitoring programs with stable pointings.
    elif user_config['star_selection_strategy'] == 'stars_per_frame':
        logger.info(f'config star selection strategy: combined largest footprint of the frames')
        largest_footprint, _ = load_combined_footprint_from_db(frames_hash)
        region_type = 'polygon'
        query_footprint = largest_footprint['coordinates'][0]
        # then, we must fall back to using stars selected in each individual frame.
        # here, we will query a larger footprint so that we have options in each
        # individual frame.
    elif user_config['star_selection_strategy'] == 'ROI_disk':
        logger.info(f'config star selection strategy: in a disk around the ROI.')
        center = user_config['ROI_ra_deg'], user_config['ROI_dec_deg']
        radius = user_config['ROI_disk_radius_arcseconds'] / 3600.0
        region_type = 'circle'
        query_footprint = {'center': center, 'radius': radius}
    else:
        raise RuntimeError("Not an agreed upon strategy for star selection:", user_config['star_selection_strategy'])

    kwargs_query = {
        'release': 'dr3',
        'astrometric_excess_noise_max': user_config['star_max_astrometric_excess_noise'],
        'gmag_range': (user_config['star_min_gmag'], user_config['star_max_gmag']),
        'max_phot_g_mean_flux_error': user_config['star_max_phot_g_mean_flux_error']
    }
    logging.info(f'Querying stars with the following parameters: {kwargs_query}')

    stars_table = find_gaia_stars(region_type, query_footprint, **kwargs_query)

    message = f"Too few stars compared to the config criterion! Only {len(stars_table)} stars available."
    enough_stars = len(stars_table) >= user_config['min_number_stars']
    if not enough_stars:
        logging.error(message + ' Force stopping.')
    assert enough_stars, message

    columns = ['combined_footprint_hash', 'name', 'ra', 'dec', 'gmag', 'rmag', 'bmag', 'pmra', 'pmdec', 'ref_epoch',
               'gaia_id', 'distance_to_roi_arcsec']
    insert_query = f"INSERT INTO stars ({', '.join(columns)}) VALUES ({', '.join(len(columns)*['?'])})"
    stars_coord = SkyCoord(ra=stars_table['ra'], dec=stars_table['dec'])
    stars_table['distance_to_roi'] = stars_coord.separation(user_config['ROI_SkyCoord']).arcsecond
    # we do not want the ROI itself as a reference:
    stars_table = stars_table[stars_table['distance_to_roi'] > user_config['ROI_size']]
    stars_table.sort('distance_to_roi')
    # add a friendly name to each star (a, b, c, ....)
    stars_table['name'] = generate_star_names(len(stars_table))
    for star in stars_table:
        # loads of floats because float32 does weird stuff to sqlite? not sure, but explicit casting fixes issues.
        star_data = (frames_hash, star['name'], float(star['ra']), float(star['dec']), float(star['phot_g_mean_mag']),
                     float(star['phot_rp_mean_mag']),  float(star['phot_bp_mean_mag']),
                     float(star['pmra']), float(star['pmdec']), float(star['ref_epoch']), int(star['source_id']),
                     star['distance_to_roi'])
        execute_sqlite_query(insert_query, params=star_data, is_select=False)
    logger.info('Calculating which star is in which frame.')
    populate_stars_in_frames()
    # let us also make a plot of how the gaia stars we queried are distributed within our footprint.
    query = """
    SELECT frames.id, footprints.polygon
    FROM footprints 
    JOIN frames ON footprints.frame_id = frames.id 
    WHERE frames.eliminated != 1;
    """
    results = execute_sqlite_query(query)
    polygon_list = [np.array(json.loads(result[1])) for result in results]
    save_path = user_config['plots_dir'] / 'footprints_with_gaia_stars.jpg'
    plot_footprints_with_stars(footprint_arrays=polygon_list, stars=stars_table.to_pandas(), save_path=save_path)
    logger.info(f'Plot of the queried reference Gaia stars saved at {save_path}.')



import numpy as np
from copy import deepcopy

from starred.deconvolution.loss import Loss
from starred.optim.optimization import Optimizer
from starred.deconvolution.parameters import ParametersDeconv
from starred.optim.inference_base import FisherCovariance


def get_flux_uncertainties(kwargs, kwargs_up, kwargs_down, data, noisemap, model):
    """
    Assuming the other parameters well constrained, this estimates the uncertainty on the flux.
    Args:
        kwargs: optimized starred parameters
        kwargs_up: upper bounds
        kwargs_down: lower bounds
        data: 3d array, the data
        noisemap: 3d array, same shape as data
        model: the model instance
    Returns:
        an array, the same shape as kwargs['kwargs_analytic']['a']: one uncertainty per flux.

    """
    kwargs_fixed = deepcopy(kwargs)
    del kwargs_fixed['kwargs_analytic']['a']

    parameters = ParametersDeconv(kwargs_init=kwargs,
                                  kwargs_fixed=kwargs_fixed,
                                  kwargs_up=kwargs_up,
                                  kwargs_down=kwargs_down)
    loss = Loss(data, model, parameters, noisemap ** 2,
                regularization_terms='l1_starlet')
    optim = Optimizer(loss, parameters, method='l-bfgs-b')
    optim.minimize(maxiter=10)

    fish = FisherCovariance(parameters, optim, diagonal_only=True)
    fish.compute_fisher_information()
    k_errs = fish.get_kwargs_sigma()
    return np.array(k_errs['kwargs_analytic']['a'])



import numpy as np
from astropy.table import Table
import sep
from astropy.io import fits
from ..plotting.sources_plotting import plot_sources


def extract_stars(image_background_subtracted, variance_map, detection_threshold=3, min_area=10,
                  debug_plot_path=None):
    """
    Extract star positions from an image using SEP (Source Extractor as a Python library).

    Parameters:
    image_background_subtracted: image contained in numpy 2d array, with background subtracted!
    variance_map: image, map of variance.
    detection_threshold: float, in units of sigma, default 3
    min_area: int, min number of pixels in detection for it to be considered, default 10

    Returns:
    astropy.table.Table: Table of detected sources.
    """

    objects = sep.extract(data=image_background_subtracted,
                          thresh=detection_threshold,
                          var=variance_map,
                          minarea=min_area)

    sources = Table()
    for col in objects.dtype.names:
        sources[col] = objects[col]

    # just to stick to the daostarfinder way
    sources['xcentroid'] = sources['x']
    sources['ycentroid'] = sources['y']

    # remove the weirdly elongated ones -- probably most stars, and we care about stars the most
    elongation = sources['a'] / sources['b']
    med_el = np.median(elongation)
    std_el = np.std(elongation)
    sources['elongation'] = elongation
    sources = sources[sources['elongation'] < med_el + 3*std_el]

    # define some FWHM quantity
    sources['FWHM'] = 2 * (np.log(2) * (sources['a']**2 + sources['b']**2))**0.5
    # and ellipticity
    sources['ellipticity'] = 1 - sources['b'] / sources['a']

    # brightest first
    sources.sort('flux', reverse=True)

    if debug_plot_path is not None:
        debug_plot_path.parent.mkdir(exist_ok=True)
        plot_sources(sources=sources, image=image_background_subtracted, save_path=debug_plot_path)

    return sources


def extract_sources_from_sky_sub_image(image_path, sources_path, detection_threshold, min_area,
                                       exptime, background_rms_electron_per_second, debug_plot_path):
    """
      Not used in the main pipeline but can be useful. Given an image, extracts
      sources, and save them in sources_path
    Args:
        image_path:  path, str: fits file containing an image.
        sources_path:  path, str: where the sources are saved
        detection_threshold: float, significance for acceptance
        min_area: int, minimum number of pixels above threshold
        exptime: float, exposure time of the frame (used to calculate a variance map,
                 because the pipeline stores the frames in e- / second)
        background_rms_electron_per_second: float, scatter in the background in e- / second.
                                            Used to calculate the variance map as well.
        debug_plot_path: where we (potentially) save a plot of our sources.
    Returns:
        Nothing
    """
    image_electrons = exptime * fits.getdata(image_path).astype(float)
    background_rms_electrons = exptime * background_rms_electron_per_second

    variance_map = background_rms_electrons**2 + np.abs(image_electrons)

    sources = extract_stars(image_background_subtracted=image_electrons,
                            detection_threshold=detection_threshold,
                            variance_map=variance_map,
                            min_area=min_area,
                            debug_plot_path=debug_plot_path)

    sources.write(sources_path, overwrite=True)




import importlib.util
import os

from .user_config import get_user_config


def load_custom_header_parser():
    """
    dynamically load the 'parse_header' function from the user-defined file
    located at '$workdir/header_parser/header_parser.py'.
    """
    workspace_dir = get_user_config()['workdir']

    file_path = os.path.join(workspace_dir, 'header_parser', 'parse_header.py')
    module_name = 'header_parser'

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    try:
        parse_header_function = getattr(module, 'parse_header')
        return parse_header_function
    except AttributeError:
        raise ImportError("The function 'parse_header' was not found in 'header_parser.py'.")



import numpy as np
import sqlite3
from pathlib import Path
from astropy.io import fits
from astropy.wcs import WCS
from time import sleep
import logging

from .background_estimation import subtract_background
from .star_extraction import extract_stars
from .frame_characterization import ephemeris, estimate_seeing
from ..structure.user_header_parser import load_custom_header_parser


def process_new_frame(fits_file, user_config):
    """
        crops, transforms to e-/s and writes the result to our workdir.

        fits_file: path or str to a fits file.

        :param fits_file: path
        :param user_config: dictionary containing user config
        :param logger: a logger.
    """
    trim_vertical = user_config.get('trim_vertical', 0)
    trim_horizontal = user_config.get('trim_horizontal', 0)
    copied_image_relpath = Path('frames') / f"{fits_file.stem}.fits"
    logger = logging.getLogger("lightcurver.importation")

    try:
        with fits.open(str(fits_file), mode='readonly', ignore_missing_end=False, memmap=True) as hdu:
            logger.info(f'  Importing {fits_file}.')
            hdu_index = 1 if len(hdu) > 1 else 0
            data_shape = hdu[hdu_index].data.shape
            cutout_data = hdu[hdu_index].section[
                          trim_vertical:data_shape[0] - trim_vertical,
                          trim_horizontal:data_shape[1] - trim_horizontal
                          ]
            header = hdu[hdu_index].header
            logger.info(f"Read fits file at {fits_file}")
    except ValueError:
        # then we have some memmap problems, if  bzero, bscale, or blank are in header.
        with fits.open(str(fits_file), mode='readonly') as hdu:
            logger.info(f'  Importing {fits_file}.')
            hdu_index = 1 if len(hdu) > 1 else 0
            data_shape = hdu[hdu_index].data.shape
            cutout_data = hdu[hdu_index].data[
                          trim_vertical:data_shape[0] - trim_vertical,
                          trim_horizontal:data_shape[1] - trim_horizontal
                          ]
            header = hdu[hdu_index].header
            logger.warning(f"Read fits file at {fits_file}, could not use memmap!!")
    cutout_data = cutout_data.astype(float)
    wcs = WCS(header)
    # so we cropped our data, thus we need to change the CRPIX of our WCS
    if wcs.is_celestial:
        wcs.wcs.crpix[0] -= trim_horizontal
        wcs.wcs.crpix[1] -= trim_vertical

    header['BUNIT'] = "ELPERSEC"

    # the user defines their header parser, returning a dictionary with {'mjd':, 'gain':, 'filter':, 'exptime':}
    # it needs be located at $workdir/header_parser/parse_header.py
    # the function needs be called parse_header, and it has to accept a fits header as argument and return the
    # dictionary above.
    header_parser_function = load_custom_header_parser()
    mjd_gain_filter_exptime_dict = header_parser_function(header)
    logger.info(f'  file {fits_file}: parsed header.')
    cutout_data *= mjd_gain_filter_exptime_dict['gain'] / mjd_gain_filter_exptime_dict['exptime']
    # unit: electron per second

    # ok, now subtract sky!
    cutout_data_sub, bkg = subtract_background(cutout_data,
                                               mask_sources_first=user_config['mask_sources_before_background'],
                                               n_boxes=user_config['background_estimation_n_boxes'])
    sky_level_electron_per_second = float(bkg.globalback)
    background_rms_electron_per_second = float(bkg.globalrms)
    logger.info(f'  file {fits_file}: background estimated.')

    # before we write, let's keep as much as we can from our previous header
    new_header = wcs.to_header()
    # then copy non-WCS entries from the original header
    for key in header:
        if key not in new_header and not key.startswith('WCSAXES') and not key.startswith(
                'CRPIX') and not key.startswith('CRVAL') and not key.startswith('CDELT') and not key.startswith(
                'CTYPE') and not key.startswith('CUNIT') and not key.startswith('CD') and key not in ['COMMENT',
                                                                                                      'HISTORY']:
            if key.strip():  # some headers are weird
                new_header[key] = header[key]
    # now we can write the file
    write_file = user_config['workdir'] / copied_image_relpath
    logger.info(f'    Writing file: {write_file}')
    fits.writeto(write_file, cutout_data_sub.astype(np.float32),
                 header=new_header, overwrite=True)
    # and find sources
    # (do plots if toggle set)
    do_plot = user_config.get('source_extraction_do_plots', False)
    plot_path = user_config['plots_dir'] / 'source_extraction' / f'{fits_file.stem}.jpg' if do_plot else None
    # we need a proper noise map here to correctly detect our sources.
    exptime = mjd_gain_filter_exptime_dict['exptime']
    cutout_data_sub_el = cutout_data_sub * exptime
    background_rms = bkg.globalrms * exptime
    variance_map = background_rms**2 + np.abs(cutout_data_sub_el)
    sources_table = extract_stars(image_background_subtracted=cutout_data_sub,
                                  variance_map=variance_map / exptime**2,  # scale back to (e-/s)^2
                                  detection_threshold=user_config.get('source_extraction_threshold', 3),
                                  min_area=user_config.get('source_extraction_min_area', 10),
                                  debug_plot_path=plot_path)
    # saving the sources in the same dir as the frame itself
    sources_file_filename = f"{copied_image_relpath.stem}_sources{copied_image_relpath.suffix}"
    sources_file_relpath = copied_image_relpath.parent / sources_file_filename
    sources_table.write(user_config['workdir'] / sources_file_relpath, format='fits', overwrite=True)

    seeing_pixels = estimate_seeing(sources_table)
    ellipticity = np.nanmedian(sources_table['ellipticity'])
    logger.info(
        f"Extracted {len(sources_table)} sources from {fits_file}. "
        f"Seeing pixels: {seeing_pixels:.02f}. "
        f"Ellipticity: {ellipticity:.02f}."
    )

    if 'telescope' in user_config:
        eph_dict = ephemeris(mjd=mjd_gain_filter_exptime_dict['mjd'],
                             ra_object=user_config['ROI_ra_deg'],
                             dec_object=user_config['ROI_dec_deg'],
                             telescope_longitude=user_config['telescope']['longitude'],
                             telescope_latitude=user_config['telescope']['latitude'],
                             telescope_elevation=user_config['telescope']['elevation'])
        telescope_information = {k: v for k, v in user_config['telescope'].items()}
        if eph_dict['weird_astro_conditions']:
            logger.warning(
                f'Ephemeris: weird for this frame ({fits_file}), please inspect and make sure your '
                'config (such as telescope location and ROI coordinates) and fits MJD are all correct. '
                f'Ephemeris dictionary for reference: {eph_dict}'
            )
    else:
        logger.warning(
            'Telescope information not provided in user config. No Ephemeris calculated. '
            'Consider adding them as this can help catch mistakes.'
        )
        telescope_information = None
        eph_dict = None

    # now we're ready for registering our new frame!
    add_frame_to_database(original_image_path=fits_file,
                          copied_image_relpath=copied_image_relpath,
                          sources_relpath=sources_file_relpath,
                          mjd=mjd_gain_filter_exptime_dict['mjd'],
                          gain=mjd_gain_filter_exptime_dict['gain'],
                          sky_level_electron_per_second=sky_level_electron_per_second,
                          background_rms_electron_per_second=background_rms_electron_per_second,
                          filter=mjd_gain_filter_exptime_dict['filter'],
                          exptime=mjd_gain_filter_exptime_dict['exptime'],
                          seeing_pixels=seeing_pixels,
                          ellipticity=ellipticity,
                          ephemeris_dictionary=eph_dict,
                          telescope_information=telescope_information,
                          user_config=user_config)
    return header


def add_frame_to_database(original_image_path, copied_image_relpath, sources_relpath,
                          mjd, gain,
                          sky_level_electron_per_second, background_rms_electron_per_second,
                          filter, exptime,
                          seeing_pixels, ellipticity, user_config,
                          telescope_information=None,
                          ephemeris_dictionary=None):
    """
    Adding our new image frame to our sqlite3 database. We will use the table "frames".

    :param original_image_path: Path to the original image
    :param copied_image_relpath: Path where the image was copied, relative to workdir
    :param sources_relpath: filename of the file of sources (fits table) as extracted by sep, relative to workdir
    :param mjd: float, mjd of frame
    :param gain: float
    :param sky_level_electron_per_second: float
    :param background_rms_electron_per_second: float
    :param filter: string, filter of the observations
    :param exptime: float, exposure time
    :param seeing_pixels: Seeing value measured for this frame, float.
    :param ellipticity: ellipticity of the psf, calculated as 1 - b/a
    :param user_config: dictionary containing the user config
    :param telescope_information: Optional dictionary with telescope information
    :param ephemeris_dictionary: dictionary as returned by the ephemeris function of frame_characterization.
    :return: None
    """
    columns = ['original_image_path', 'image_relpath', 'sources_relpath',
               'seeing_pixels', 'mjd', 'gain', 'sky_level_electron_per_second', 'background_rms_electron_per_second',
               'filter', 'exptime', 'ellipticity']

    values = [str(original_image_path), str(copied_image_relpath), str(sources_relpath),
              seeing_pixels, mjd, gain,
              sky_level_electron_per_second, background_rms_electron_per_second,
              filter, exptime, ellipticity]

    # if telescope information, add it to the columns and values
    if telescope_information is not None:
        for key, value in telescope_information.items():
            columns.append(f"telescope_{key}")
            values.append(value)

    if ephemeris_dictionary is not None:
        columns.append('airmass')
        values.append(float(ephemeris_dictionary['target_info']['airmass']))
        columns.append('degrees_to_moon')
        values.append(ephemeris_dictionary['moon_info']['distance_deg'])
        columns.append('moon_phase')
        values.append(ephemeris_dictionary['moon_info']['illumination'])
        columns.append('sun_altitude')
        values.append(ephemeris_dictionary['sun_info']['altitude_deg'])
        columns.append('azimuth')
        values.append(ephemeris_dictionary['target_info']['azimuth_deg'])
        columns.append('altitude')
        values.append(ephemeris_dictionary['target_info']['altitude_deg'])

    # we'll return this whole thing at the end for potential further use
    frame_info = {key: value for key, value in zip(columns, values)}

    # construct the SQL query based on the columns
    column_names = ', '.join(columns)
    placeholders = ', '.join(['?'] * len(columns))
    query = f'INSERT INTO frames ({column_names}) VALUES ({placeholders})'

    # the query
    try:
        inserted = False
        while not inserted:
            # We are being very careful, sqlite3 databases are not made for parallel writing.
            # Works well so long as the storage (on which the database is) is fairly responsive.
            # Local hard drives and SSDs all work, virtual storage volumes as well.
            # Only case of failure: an SSD I have that clearly has problems (response time 5-15 seconds for
            # touching a file)
            try:
                conn = sqlite3.connect(user_config['database_path'], timeout=2.5)
                cursor = conn.cursor()
                # cursor.execute("PRAGMA journal_mode=WAL")
                # cursor.execute("PRAGMA busy_timeout=2500")
                cursor.execute(query, values)
                conn.commit()
                inserted = True
                conn.close()
            except sqlite3.OperationalError:
                print('database locked, waiting')
                sleep(np.random.uniform(0, 0.5))  # retry after random wait
            finally:
                try:
                    conn.close()
                except:
                    pass
    except sqlite3.IntegrityError as IntE:
        logger = logging.getLogger("lightcurver.importation")
        logger.error(
            "Error: we might be inserting an already existing image again in the database. "
            "You most likely overwrote an existing file already, leaving us in an inconsistent state. "
            f"This should not have happened and will take manual fixing. Here was the original error: {IntE} "
            "We will raise the error again so you can check the traceback. "
            "If you see a UNIQUE violation on the image_relpath column, then the above is indeed what happened."
            "Force stopping pipeline."
        )
        raise IntE
    return frame_info




import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib as mpl
import pandas as pd

from ..structure.database import execute_sqlite_query
from ..utilities.chi2_selector import get_chi2_bounds


def plot_normalized_star_curves(combined_footprint_hash, save_path=None):
    """
    Given our config params (in particular chi2 rejection criterion), plot the fluxes of the stars that were
    used to build the normalization coefficient of the frames, normalized by this very coefficient.
    Args:
        combined_footprint_hash: footprint hash that we are working with.
        save_path: path or str, where we save the plot.

    Returns:
        None
    """
    # doing this because we set the stylesheet to dark for other plots, and we don't want it here (hard to read)
    mpl.rcParams.update(mpl.rcParamsDefault)

    # query the coefficients and fluxes
    fluxes_fit_chi2_min, fluxes_fit_chi2_max = get_chi2_bounds(psf_or_fluxes='fluxes')

    query_norm = """
    SELECT 
       nc.*, f.mjd
    FROM 
       normalization_coefficients nc
    JOIN
       frames f ON nc.frame_id = f.id
    WHERE   
       coefficient > coefficient_uncertainty
    AND
       combined_footprint_hash = ?
    """
    df_norm = execute_sqlite_query(query=query_norm,
                                   params=(combined_footprint_hash,),
                                   use_pandas=True, is_select=True)
    query_fluxes = """
    SELECT 
       sff.*, s.name AS name, s.gmag as gmag
    FROM 
       star_flux_in_frame sff
    JOIN 
       stars s ON sff.star_gaia_id = s.gaia_id
    WHERE 
       sff.chi2 > ? 
    AND
       sff.chi2 < ?
    AND
       sff.combined_footprint_hash = ?
    """
    df_fluxes = execute_sqlite_query(query=query_fluxes,
                                     params=(fluxes_fit_chi2_min, fluxes_fit_chi2_max, combined_footprint_hash,),
                                     use_pandas=True, is_select=True)
    df = pd.merge(df_norm, df_fluxes, on=['frame_id'])

    # normalize the fluxes
    df['normalized_flux'] = df['flux'] / df['coefficient']
    df['normalized_uncertainty'] = df['flux_uncertainty'] / df['coefficient']

    # ok, prepare plotting
    star_names = df['name'].unique()
    plot_size = 3.5
    n_rows = len(star_names) + 1

    fig, axs = plt.subplots(n_rows, 1, figsize=(2 * plot_size, n_rows * plot_size), sharex=True)
    axs = axs.flatten()

    # norm color scale across all stars based on chi2 values
    norm = colors.Normalize(vmin=fluxes_fit_chi2_min, vmax=fluxes_fit_chi2_max)
    cmap = plt.get_cmap('coolwarm')

    # norm coefficient plot
    axs[0].errorbar(df_norm['mjd'], df_norm['coefficient'],
                    yerr=df_norm['coefficient_uncertainty'],
                    fmt='.', ms=0,
                    ecolor='gray', alpha=0.8, zorder=-1000, elinewidth=0.8)
    sc = axs[0].scatter(df_norm['mjd'], df_norm['coefficient'], s=5, edgecolor='none', color='red')
    axs[0].set_ylabel('Normalization Coefficient')
    axs[0].grid(True)
    cl = fig.colorbar(sc, ax=axs[0], label='chi2')

    # norm star fluxes
    for i, name in enumerate(sorted(star_names)):
        ax = axs[i + 1]
        star_data = df[df['name'] == name]
        medflux = star_data['normalized_flux'].median()
        err = ((star_data['normalized_uncertainty'] / medflux)**2 + (star_data['coefficient_uncertainty'])**2)**0.5
        ax.errorbar(star_data['mjd'], star_data['normalized_flux'] / medflux,
                    yerr=err, fmt='.', ms=0, ecolor='gray', alpha=0.5, zorder=-1000, elinewidth=0.5)
        sc = ax.scatter(star_data['mjd'], star_data['normalized_flux'] / medflux,
                        c=star_data['chi2'], cmap=cmap, norm=norm, s=10,
                        edgecolor='none', label=f"Star {name} (gmag: {star_data['gmag'].unique()[0]:.1f})")
        ax.set_ylabel('Normalized flux / global median')
        ax.set_ylim((0.9, 1.1))
        ax.grid(True)
        ax.legend()

        fig.colorbar(sc, ax=ax, label='chi2')

    # time label
    axs[-1].set_xlabel('MJD')

    plt.tight_layout()
    cl.set_label('')
    cl.set_ticks([])

    if save_path is not None:
        plt.savefig(save_path, pad_inches=0, bbox_inches='tight', dpi=150)
        plt.close()
    else:
        return fig, axs



import numpy as np
from pathlib import Path
import h5py
import sep
import logging
from time import time
from starred.procedures.psf_routines import build_psf

from ..structure.database import select_stars_for_a_frame, execute_sqlite_query, get_pandas
from ..structure.user_config import get_user_config
from ..plotting.psf_plotting import plot_psf_diagnostic
from ..utilities.footprint import get_combined_footprint_hash


def check_psf_exists(frame_id, psf_ref, combined_footprint_hash):
    """
    just checks if the PSF in question has already been modelled. (for same frame ID, of the specific combined footprint
    and same name constructed from the stars used in the model).
    Args:
        frame_id: integer, database id of the frame
        psf_ref: string, e.g. 'psf_abcd' for a psf model using stars labelled a,b,c,d.
        combined_footprint_hash: integer, the hash of the combined footprint at hand.

    Returns:
        bool: whether such a psf row is in the PSFs table or not.

    """
    query = "SELECT 1 FROM PSFs WHERE frame_id = ? AND psf_ref = ? and combined_footprint_hash = ?"
    params = (frame_id, psf_ref, combined_footprint_hash)
    result = execute_sqlite_query(query, params)
    return len(result) > 0


def mask_surrounding_stars(data, noisemap):
    """
    masks any object that is not the central one. (pixel ok = True, pixel masked = False)
    Args:
        data: 2d array
        noisemap:  2d array

    Returns:
        mask (2d array, one if good pixel, zero if masked)
    """
    objects, seg_map = sep.extract(data, thresh=3., err=noisemap, minarea=15, segmentation_map=True,
                                   deblend_cont=0.001)

    mask = np.ones_like(seg_map, dtype=bool)
    if len(objects) == 0:
        return mask
    center_y = (data.shape[0] - 1) / 2.0
    center_x = (data.shape[1] - 1) / 2.0
    distances = np.sqrt((objects['x'] - center_x) ** 2 + (objects['y'] - center_y) ** 2)

    central_star_index = np.argmin(distances)

    for i in range(0, len(objects)):
        if i != central_star_index:
            mask[seg_map == i + 1] = False

    return mask


def model_all_psfs():
    """
    This is called by the workflow manager.
    It
    - queries all the frames that should have a PSF model built for
    - given the combined footprint at hand, selects the available stars for each frame
    - does additional checks on the cutouts of the stars (such as removing stars with too many masked pixels)
    - builds the psf with starred
    - logs results into the hdf5 file and updates the database.
    Returns:
        Nothing
    """
    logger = logging.getLogger('lightcurver.psf_modelling')
    user_config = get_user_config()
    stars_to_use = user_config['stars_to_use_psf']

    # where we'll save our stamps
    regions_file = user_config['regions_path']

    # query frames
    frames = get_pandas(columns=['id', 'image_relpath', 'exptime', 'mjd', 'seeing_pixels', 'pixel_scale'],
                        conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames['id'].to_list())

    logger.info(f'Will build the PSFs of (potentially) {len(frames)} frames, storing result in file {regions_file}.')
    # for each frame, will check if the PSF was already built.

    for i, frame in frames.iterrows():
        t0 = time()
        stars = select_stars_for_a_frame(frame_id=frame['id'],
                                         combined_footprint_hash=combined_footprint_hash,
                                         stars_to_use=stars_to_use)
        stars.sort_values(by=['name'])
        if len(stars) == 0:
            # we simply do not build a PSF. the frame will not be considered in the joint queries later.
            logger.warning(f"The frame with id {frame['id']} does not have available reference stars. Skipping.")
            continue
        psf_ref = 'psf_' + ''.join(sorted(stars['name']))

        # check so we don't redo for nothing
        if check_psf_exists(frame['id'], psf_ref, combined_footprint_hash) and not user_config['redo_psf']:
            logger.info(
                f"The frame with id {frame['id']} already has a PSF (ref {psf_ref}), redo flag not set. Skipping."
            )
            continue

        # get the cutouts
        with h5py.File(regions_file, 'r') as f:
            data_group = f[f"{frame['image_relpath']}/data"]
            noisemap_group = f[f"{frame['image_relpath']}/noisemap"]
            mask_group = f[f"{frame['image_relpath']}/cosmicsmask"]
            datas = np.array([data_group[gaia_id][...] for gaia_id in list(stars['gaia_id'])])
            noisemaps = np.array([noisemap_group[name][...] for name in list(stars['gaia_id'])])
            cosmics_masks = np.array([mask_group[name][...] for name in list(stars['gaia_id'])]).astype(bool)
            # invert because the cosmics are marked as True, but we want the healthy pixels to be marked as True:
            cosmics_masks = ~cosmics_masks
        # now we'll prepare automatic masks (masking other objects in the field)
        automatic_masks = []
        for data, noisemap in zip(datas, noisemaps):
            mask = mask_surrounding_stars(data, noisemap)
            automatic_masks.append(mask)
        automatic_masks = np.array(automatic_masks)
        # now both cosmics_masks and automatic_masks have 0 for bad pixels, and 1 for good pixels.
        masks = cosmics_masks * automatic_masks
        isnan = np.where(np.isnan(datas)*np.isnan(noisemaps))
        # before eliminating entire slices, 'mask' the NaNs
        datas[isnan] = 0.
        noisemaps[isnan] = 1.0
        masks[isnan] = False
        n_stars_before = len(datas)
        # now eliminate slices, which changes the shapes.
        # we'll delete the slices with too many masked pixels (we mask anything with more than 40% masked)
        mask_threshold_fraction = 0.4
        total_elements_per_slice = datas.shape[1] * datas.shape[2]
        masked_counts_per_slice = np.sum(~masks, axis=(1, 2))
        slices_to_remove = masked_counts_per_slice > (mask_threshold_fraction * total_elements_per_slice)
        slices_to_keep = ~slices_to_remove
        datas = datas[slices_to_keep]
        noisemaps = noisemaps[slices_to_keep]
        masks = masks[slices_to_keep]
        names = list(stars['name'][slices_to_keep])
        n_stars_after = len(datas)
        if n_stars_after == 0:
            # no psf model for this one.
            logger.warning(
                f"The frame with id {frame['id']} had {n_stars_before} reference stars, "
                "but none could be used due to too many masked pixels. Skipping."
            )
            continue

        # we set the initial guess for the position of the star to the center (guess_method thing)
        # because we are confident that is where the star will be (plate solving + gaia proper motions)
        result = build_psf(datas, noisemaps, subsampling_factor=user_config['subsampling_factor'],
                           n_iter_analytic=user_config['psf_n_iter_analytic'],
                           n_iter_adabelief=user_config['psf_n_iter_pixels'],
                           masks=masks,
                           guess_method_star_position='center')
        psf_plots_dir = user_config['plots_dir'] / 'PSFs' / str(combined_footprint_hash)
        psf_plots_dir.mkdir(exist_ok=True, parents=True)
        frame_name = Path(frame['image_relpath']).stem
        seeing = frame['seeing_pixels'] * frame['pixel_scale']
        loss_history = result['adabelief_extra_fields']['loss_history']
        plot_psf_diagnostic(datas=datas, noisemaps=noisemaps, residuals=result['residuals'],
                            full_psf=result['full_psf'],
                            loss_curve=loss_history,
                            masks=masks, names=names,
                            diagnostic_text=f"{frame_name}\nseeing: {seeing:.02f}",
                            save_path=psf_plots_dir / f"{frame['id']}_{frame_name}.jpg")

        # now we can do the bookkeeping stuff
        with h5py.File(regions_file, 'r+') as f:
            # if already there, delete it before replacing.
            frame_group = f[frame['image_relpath']]
            if psf_ref in frame_group.keys():
                del frame_group[psf_ref]
            psf_group = frame_group.create_group(psf_ref)
            psf_group['narrow_psf'] = np.array(result['narrow_psf'])
            psf_group['full_psf'] = np.array(result['full_psf'])
            psf_group['subsampling_factor'] = np.array([user_config['subsampling_factor']])

        # and update the database.
        loss_index = int(0.9 * np.array(loss_history).size)
        initial_change = np.nanmax(loss_history[:loss_index]) - np.nanmin(loss_history[:loss_index])
        end_change = np.nanmax(loss_history[loss_index:]) - np.nanmin(loss_history[loss_index:])
        relative_loss_differential = float(end_change / initial_change)
        insert_params = (frame['id'], float(result['chi2']), relative_loss_differential, psf_ref,
                         combined_footprint_hash, user_config['subsampling_factor'])
        insert_query = "REPLACE INTO PSFs "
        insert_query += "(frame_id, chi2, relative_loss_differential, psf_ref, combined_footprint_hash, subsampling_factor) "
        insert_query += f"VALUES ({','.join(['?'] * len(insert_params))})"

        execute_sqlite_query(insert_query, insert_params, is_select=False)

        time_taken = time() - t0
        logger.info(
            f"PSF built for frame with id {frame['id']}. "
            f"The reference is {psf_ref}, that is {n_stars_before} stars available, and {n_stars_after} actually used "
            "after filtering of masked pixels. "
            f"The reduced chi2 is {result['chi2']:.02f}, and it took us {time_taken:.01f} seconds to complete everything. "
        )





from shapely.geometry import Point, Polygon
from shapely import intersection
import json
import sqlite3
import numpy as np

from ..structure.user_config import get_user_config


def populate_stars_in_frames():
    """
    Populates the stars_in_frames table by checking which stars fall within each frame's footprint.
    """

    user_config = get_user_config()
    # not using execute_sqlite_query, so we do not open and close the database on million times.
    conn = sqlite3.connect(user_config['database_path'])
    cursor = conn.cursor()

    # load the footprints and stars
    cursor.execute("""SELECT frame_id, polygon FROM footprints""")
    frame_footprints = cursor.fetchall()
    cursor.execute("""SELECT gaia_id, ra, dec, combined_footprint_hash FROM stars""")
    stars = cursor.fetchall()

    for frame_id, footprint_str in frame_footprints:
        # assume gnomonic projection, let's just treat our footprint as flat.
        # should be good enough.
        footprint_polygon = Polygon(json.loads(footprint_str))
        # we don't want to select stars too close to the edges, too many problems.
        # to address this, we'll shrink our polygon by a margin.

        # get the edges:
        x, y = footprint_polygon.exterior.xy
        # the mean declination:
        mean_dec = np.nanmean(y)
        margin_degrees = 4. / 3600  # enforce a margin of 15 arcseconds.
        # just de-projecting the margin for RA:
        ra_margin = margin_degrees / np.cos(np.radians(mean_dec))

        # I do not know how to shrink y polygon by different amounts in different directions ...
        # no worries, we'll just translate the polygon by the amount in a cross-like pattern,
        # then take the intersection to get our 'shrunk' polygon.
        translated_polygons = []
        for translation in ([1, 0], [-1, 0], [0, -1], [0, 1]):
            adjusted_vertices = []
            for ra, dec in zip(x, y):
                adjusted_ra = ra + translation[0] * ra_margin
                adjusted_dec = dec + translation[1] * margin_degrees
                adjusted_vertices.append((adjusted_ra, adjusted_dec))
            translated_polygons.append(Polygon(adjusted_vertices))

        # do the intersection of our translations to get the reduced footprint:
        shrunk_footprint_polygon = translated_polygons[0]
        for trans in translated_polygons[1:]:
            shrunk_footprint_polygon = intersection(shrunk_footprint_polygon, trans)

        # and now we check!
        for star_id, ra, dec, combined_footprint_hash in stars:
            star_point = Point(ra, dec)
            if star_point.within(shrunk_footprint_polygon):
                try:
                    cursor.execute("""INSERT INTO stars_in_frames (frame_id, star_gaia_id, combined_footprint_hash) 
                                      VALUES (?, ?, ?)""", (frame_id, star_id, combined_footprint_hash))
                except sqlite3.IntegrityError:
                    # handles cases where the same star-frame relationship might already be in the table
                    continue

    conn.commit()
    conn.close()




import matplotlib.pyplot as plt
import numpy as np


def plot_joint_deconv_diagnostic(datas, noisemaps, residuals, loss_curve, chi2_per_frame,
                                 save_path=None):
    """
    a diagnostic of how well a joint deconv went. We will plot a stack of the data, and a stack of
    the residuals. We will show how the loss behaved during optimization.
    It is important to avoid systematics, as this is the core of this package: doing precise photometry,
    so convergence when modelling the pixels is essential.
    Args:
        datas: 3d array of shape (N, nx, ny): N stars, one slice per star.
        noisemaps: same as above but for the noisemaps
        residuals: same as above but for the residuals
        loss_curve: 1D array containing the evolution of the loss during optimization
        chi2_per_frame: 1D array, the chi2 value of the fit, one per slice.
        save_path: optional, string or path, where to save the plot.

    Returns: None

    """
    plt.style.use('dark_background')
    cmap = 'viridis'
    cmap_residuals = 'coolwarm'
    text_color = 'red'
    text_size = 11

    data_stack = np.mean(datas, axis=0)  # mean and not median, we wanna see them outliers
    residuals_stack = np.mean(residuals, axis=0)
    rel_residuals_stack = np.mean(residuals / noisemaps, axis=0)

    sub_size = 3
    fig, ax = plt.subplots(1, 5, figsize=(4.6 * sub_size, sub_size))
    ax = ax.flatten()
    # data stack
    ax[0].imshow(data_stack, cmap=cmap, aspect='auto')
    ax[0].axis('off')
    ax[0].text(0.5, 0.01, 'Data stack',
               horizontalalignment='center',
               verticalalignment='bottom',
               transform=ax[0].transAxes,
               color=text_color, fontsize=text_size,
               weight='bold')
    # residuals stack
    ax[1].imshow(residuals_stack, cmap=cmap_residuals, aspect='auto')
    ax[1].axis('off')
    ax[1].text(0.5, 0.01, 'residuals stack',
               horizontalalignment='center',
               verticalalignment='bottom',
               transform=ax[1].transAxes,
               color=text_color, fontsize=text_size,
               weight='bold')

    # rel residuals stack
    ax[2].imshow(rel_residuals_stack, cmap=cmap_residuals, aspect='auto')
    ax[2].axis('off')
    ax[2].text(0.5, 0.01, 'rel. residuals stack',
               horizontalalignment='center',
               verticalalignment='bottom',
               transform=ax[2].transAxes,
               color=text_color, fontsize=text_size,
               weight='bold')

    # loss plot
    ax[3].plot(loss_curve, color='white')
    ax[3].text(0.5, 0.99, 'loss',
               horizontalalignment='center',
               verticalalignment='top',
               transform=ax[3].transAxes,
               color='white', fontsize=text_size,
               weight='bold')

    ax[3].axis('off')
    # chi2 plot
    ax[4].hist(chi2_per_frame, color='white', bins=len(chi2_per_frame))
    ax[4].text(0.5, 0.99, 'chi2 per frame',
               horizontalalignment='center',
               verticalalignment='top',
               transform=ax[4].transAxes,
               color='white', fontsize=text_size,
               weight='bold')
    plt.tight_layout()

    if save_path is not None:
        plt.savefig(save_path, pad_inches=0, bbox_inches='tight', dpi=130)
        plt.close()
    else:
        plt.show()



import sep
import numpy as np


def subtract_background(image, mask_sources_first=False, n_boxes=10):
    """
    subtracts an estimated smooth background from the 2d array "image".
    basic, for complex cases do your own background subtraction.
    Background subtraction is particularly sensitive if we want to achieve
    high precision on the relative zeropoints of the images.
    (we will not include a background in our deconvolutions, because it can always be
    degenerate with other sources or the object itself).

    Here we will do this in two steps if mask_sources_first is True.
    1. roughly subtract the background, extract the sources.
    2. mask the sources, estimate the background again.

    :param image: 2d numpy array
    :param mask_sources_first: bool, whether we identify sources and mask before doing background estimation.
    :param n_boxes: int, in how many boxes do we divide the side of the image for background estimation?
    :return: 2d numpy array
    """
    # so, first, estimate a background.
    box_size = np.min(image.shape) // n_boxes
    bkg = sep.Background(image, bw=box_size, bh=box_size, fw=3, fh=3)
    image_sub = image - bkg
    if not mask_sources_first:
        # that's it
        return image_sub, bkg

    # find a lot of sources.
    sources, seg_map = sep.extract(data=image_sub, var=bkg.globalrms**2, thresh=2,
                                   minarea=10, segmentation_map=True)
    # estimate again
    bkg = sep.Background(image, bw=box_size, bh=box_size, fw=3, fh=3, mask=(seg_map > 0))
    # sub again
    image_sub = image - bkg
    # and done
    return image_sub, bkg



from astroquery.gaia import Gaia


def construct_where_conditions(release, astrometric_excess_noise_max=None, gmag_range=None,
                               max_phot_g_mean_flux_error=None):
    """
    utility function for the find functions below, just checks on the conditions and returns a list of potential
    statements to us after 'WHERE'. Also takes care of formatting the table from which we will query.
    Args:
        release: string, 'dr2' or 'dr3'
        astrometric_excess_noise_max: float, default None
        gmag_range: tuple of floats (min, max), default None
        max_phot_g_mean_flux_error: float, default None

    Returns:
        list of strings containing the conditions, and string containing the name of the table to query.

    """
    assert release.lower() in ['dr2', 'dr3'], "Release must be either 'dr2' or 'dr3'"
    query_table = f"gaia{release.lower()}.gaia_source"
    where_conditions = []

    if astrometric_excess_noise_max is not None:
        where_conditions.append(f"{query_table}.astrometric_excess_noise < {astrometric_excess_noise_max}")

    if gmag_range is not None:
        where_conditions.append(f"{query_table}.phot_g_mean_mag BETWEEN {gmag_range[0]} AND {gmag_range[1]}")

    if max_phot_g_mean_flux_error is not None:
        where_conditions.append(f"{query_table}.phot_g_mean_flux_error < {max_phot_g_mean_flux_error}")

    return where_conditions, query_table


def find_gaia_stars(region_type, *args, **kwargs):
    """
    Main function to query Gaia stars based on region type (Circle or Polygon).

    :param region_type: str, 'circle' or 'polygon' to define the type of region for the query.
    :param args: Arguments passed to the specific region query function.
    :param kwargs: Keyword arguments for filtering options, passed to the specific region query function.
    """
    if region_type.lower() == 'circle':
        stars_table = find_gaia_stars_in_circle(*args, **kwargs)
    elif region_type.lower() == 'polygon':
        stars_table = find_gaia_stars_in_polygon(*args, **kwargs)
    else:
        raise ValueError("region_type must be either 'Circle' or 'Polygon'")

    # it seems that there is some variation when querying gaia? sometimes columns are capitalized, others not.
    # so, force lower:
    for name in stars_table.colnames:
        new_name = name.lower()
        stars_table.rename_column(name, new_name)
    return stars_table


def find_gaia_stars_in_circle(center_radius, release='dr3', astrometric_excess_noise_max=None, gmag_range=None,
                              max_phot_g_mean_flux_error=None):
    """
    Query Gaia stars within a circle defined by a central point and radius, with additional filtering options.

    :param center_radius: dictionary {'center': tuple(ra,dec), 'radius': radius_degrees}
    :param release: str, default 'dr3'. Either 'dr3' or 'dr2'
    :param astrometric_excess_noise_max: float, maximum allowed astrometric excess noise (None for no filter)
    :param gmag_range: tuple, magnitude range in g-band as (min_gmag, max_gmag) (None for no filter)
    :param max_phot_g_mean_flux_error: float, flux error correlates to variability. (None for no filter)

    Returns: astropy table of gaia sources
    """
    where_conditions, query_table = construct_where_conditions(release, astrometric_excess_noise_max,
                                                               gmag_range, max_phot_g_mean_flux_error)
    Gaia.MAIN_GAIA_TABLE = query_table
    Gaia.ROW_LIMIT = 3000
    # Constructing the circle condition
    c = center_radius['center']
    r = center_radius['radius']
    where_conditions.append(
        f"1=CONTAINS(POINT('ICRS', {query_table}.ra, {query_table}.dec), CIRCLE('ICRS', {c[0]}, {c[1]}, {r}))"
    )

    where_clause = " AND ".join(where_conditions)

    adql_query = f"""
    SELECT * FROM {query_table}
    """
    if where_clause:
        adql_query += f" WHERE {where_clause}"
    job = Gaia.launch_job_async(adql_query)
    result = job.get_results()
    return result


def find_gaia_stars_in_polygon(vertices, release='dr3', astrometric_excess_noise_max=None, gmag_range=None,
                               max_phot_g_mean_flux_error=None):
    """
    Query Gaia stars within a polygon defined by a list of vertices, with additional filtering options.

    :param vertices: list of tuples, each tuple contains RA and Dec in degrees [(ra1, dec1), (ra2, dec2), ..., ]
    :param release: str, default 'dr3'. Either 'dr3' or 'dr2'
    :param astrometric_excess_noise_max: float, maximum allowed astrometric excess noise (None for no filter)
    :param gmag_range: tuple, magnitude range in g-band as (min_gmag, max_gmag) (None for no filter)
    :param max_phot_g_mean_flux_error: float, flux error correlates to variability. (None for no filter)

    Returns: astropy table of gaia sources
    """
    where_conditions, query_table = construct_where_conditions(release, astrometric_excess_noise_max,
                                                               gmag_range, max_phot_g_mean_flux_error)
    Gaia.MAIN_GAIA_TABLE = query_table
    Gaia.ROW_LIMIT = 3000

    polygon_string = ', '.join([f"{vertex[0]},{vertex[1]}" for vertex in vertices])
    where_conditions.append(
        f"1=CONTAINS(POINT('ICRS', {query_table}.ra, {query_table}.dec), POLYGON('ICRS', {polygon_string}))"
    )

    where_clause = " AND ".join(where_conditions)

    adql_query = f"""
    SELECT * FROM {Gaia.MAIN_GAIA_TABLE}
    WHERE {where_clause}
    """
    job = Gaia.launch_job_async(adql_query)
    result = job.get_results()
    return result




import h5py
import numpy as np
import sqlite3
from datetime import datetime
import logging
from time import time
from starred.deconvolution.deconvolution import setup_model
from starred.deconvolution.loss import Loss
from starred.deconvolution.parameters import ParametersDeconv
from starred.optim.optimization import Optimizer

from ..structure.database import execute_sqlite_query, select_stars, select_stars_for_a_frame, get_pandas
from ..structure.user_config import get_user_config
from ..utilities.chi2_selector import get_chi2_bounds
from ..utilities.footprint import get_combined_footprint_hash
from ..utilities.starred_utilities import get_flux_uncertainties
from ..plotting.star_photometry_plotting import plot_joint_deconv_diagnostic


def do_one_deconvolution(data, noisemap, psf, subsampling_factor, n_iter=2000):
    """
    Joint 'deconvolution' of N stamps of a star (in data), with noisemap, and associated PSF at each slice.
    the subsampling factor is that used for building the psf model.
    Equivalent to PSF photometry of all slices.
    Args:
        data: numpy array (N, nx, ny) containing the epochs of the star
        noisemap:  numpy array (N, nx, ny) noisemap of the above
        psf: numpy array (N, nxx, nyy) PSF model, one per slice
        subsampling_factor: int, subsampling_factor of the psf model.
        n_iter: int, number of adabelief iterations to do. default 2000

    Returns: dictionary, containing the fluxes (ready to be used) as a 1D array, the final kwargs of the optimization,
             the chi2, and by how much we rescaled the data before optimizing. (the fluxes are already scaled back
             to the data's actual scale, but not the kwargs.)

    """
    # so, rescale

    scale = np.nanmax(data)
    data /= scale
    noisemap /= scale
    sigma_2 = noisemap**2
    # image positions: just a point source in the center.
    xs = np.array([0.])
    ys = np.array([0.])
    # initial guess for fluxes, some kind of rough "aperture"  photometry with background subtraction
    background_values = np.nanmean([
            np.nanmedian(data[:, :1, :], axis=(1, 2)),
            np.nanmedian(data[:, :, :1], axis=(1, 2)),
            np.nanmedian(data[:, -1:, :], axis=(1, 2)),
            np.nanmedian(data[:, :, -1:], axis=(1, 2))
    ])
    # just in case all nans (even though we filter them) in the borders:
    background_values = np.nan_to_num(background_values, nan=0)
    a_est = np.nansum(data, axis=(1, 2)) - data[0].size * background_values
    a_est = list(a_est)

    model, kwargs_init, kwargs_up, kwargs_down, kwargs_fixed = setup_model(data, sigma_2, psf,
                                                                           xs, ys,
                                                                           subsampling_factor,
                                                                           a_est)

    # just so we're independent of any variation in the starred building of the fixed values, let us
    # explicitly set the fixed parameters (easier to fix a missing keyword, with definite error, than a sneaky
    # change in the behaviour)
    kwargs_fixed = {
        'kwargs_analytic': {'alpha': kwargs_init['kwargs_analytic']['alpha']},
        'kwargs_background': {
            'h': kwargs_init['kwargs_background']['h'],  # very little chance that starred changes the initial h = 0.
            'mean': np.ravel([0. for _ in range(len(data))])  # we trust our background subtraction. mean = 0.
        },
        'kwargs_sersic': {},
    }

    parameters = ParametersDeconv(kwargs_init=kwargs_init,
                                  kwargs_fixed=kwargs_fixed,
                                  kwargs_up=kwargs_up,
                                  kwargs_down=kwargs_down)

    loss = Loss(data=data, deconv_class=model, param_class=parameters, sigma_2=sigma_2,
                regularization_terms='l1_starlet',
                regularization_strength_scales=1,  # not needed since no free background...
                regularization_strength_hf=1)      # but the starred interface wants us to provide them anyway

    optim = Optimizer(loss, parameters, method='adabelief')

    optimiser_optax_option = {
                'max_iterations': n_iter, 'min_iterations': None,
                'init_learning_rate': 2e-3, 'schedule_learning_rate': True,
                'restart_from_init': True, 'stop_at_loss_increase': False,
                'progress_bar': True, 'return_param_history': True
    }

    optim.minimize(**optimiser_optax_option)
    kwargs_final = parameters.best_fit_values(as_kwargs=True)
    modelled_pixels = model.model(kwargs_final)
    residuals = data - modelled_pixels
    # let's calculate a chi2 in each frame!
    chi2_per_frame = np.nansum((residuals**2 / sigma_2), axis=(1, 2)) / model.image_size**2
    chi2 = np.nanmean(chi2_per_frame)
    fluxes = scale * np.array(kwargs_final['kwargs_analytic']['a'])

    # let us calculate the uncertainties ~ equivalent to photon noise / read noise
    flux_uncertainties = get_flux_uncertainties(kwargs=kwargs_final, kwargs_down=kwargs_down, kwargs_up=kwargs_up,
                                                data=data, noisemap=noisemap, model=model)
    fluxes_uncertainties = scale * flux_uncertainties

    result = {
        'scale': scale,
        'kwargs_final': kwargs_final,
        'fluxes': fluxes,
        'fluxes_uncertainties': fluxes_uncertainties,
        'chi2': float(chi2),
        'chi2_per_frame': np.array(chi2_per_frame),
        'loss_curve': optim.loss_history,
        'residuals': scale * residuals
    }
    return result


def get_frames_for_star(combined_footprint_hash, gaia_id,
                        psf_fit_chi2_min, psf_fit_chi2_max, only_fluxless_frames=False):
    """
    Retrieves frames that include the given star, provided that those frames have a PSF with a chi2 between
    psf_fit_chi2_min and psf_fit_chi2_max. Optionally, can filter to include only frames without a flux measurement.

    :param combined_footprint_hash: int, the hash of the combined footprint we are processing.
    :param gaia_id: The Gaia ID of the star.
    :param psf_fit_chi2_min: The minimum acceptable chi2 value for the PSF fit.
    :param psf_fit_chi2_max: The maximum acceptable chi2 value for the PSF fit.
    :param only_fluxless_frames: If True, only returns frames without a flux measurement for the star. Default is False.
    :return: A list of frames that meet the criteria.
    """
    # start building the base query
    query = """
    SELECT f.*, ps.chi2, ps.psf_ref
    FROM frames f
    JOIN stars_in_frames sif ON f.id = sif.frame_id AND sif.combined_footprint_hash = ?
    """
    # if we only are selecting the frames without a flux measurement:
    if only_fluxless_frames:
        query += "LEFT JOIN star_flux_in_frame sff ON f.id = sff.frame_id AND sif.star_gaia_id = sff.star_gaia_id "
        query += "AND sif.combined_footprint_hash = sff.combined_footprint_hash\n"

    # keep building the query
    query += """
    JOIN PSFs ps ON f.id = ps.frame_id AND sif.combined_footprint_hash = ps.combined_footprint_hash
    WHERE sif.star_gaia_id = ? 
    """
    # the condition for frames without flux measurements
    if only_fluxless_frames:
        query += "AND sff.frame_id IS NULL\n"

    # finish the query with PSF chi2 conditions and existence check for appropriate PSF, selecting
    # only 1 to avoid selecting the same frame multiple times if multiple psf models are available for a given frame.
    query += """
    AND EXISTS (
        SELECT 1
        FROM PSFs ps
        WHERE f.id = ps.frame_id
        AND ps.chi2 BETWEEN ? AND ?
    )"""
    params = (combined_footprint_hash, gaia_id, psf_fit_chi2_min, psf_fit_chi2_max)

    return execute_sqlite_query(query, params, is_select=True, use_pandas=True)


def update_star_fluxes(flux_data):
    """
      just inserts (or updates) the fluxes in the database.
    Args:
        flux_data: list of tuples, like this:
                              (combined_footprint_hash, frame_id, gaia_id, flux, flux_uncertainty,
                              chi2, relative_loss_differential)
                              (but check the query below to make sure)

    Returns:
        Nothing
    """
    db_path = get_user_config()['database_path']
    with sqlite3.connect(db_path, timeout=15.0) as conn:
        cursor = conn.cursor()

        # insert query with ON CONFLICT clause for bulk update
        # upon ON CONFLICT (integrity error due to trying to insert a flux for a given star and frame again),
        # as we would if we "redo", then we do indeed erase the previous values and add the new ones.
        insert_query = """
        INSERT INTO star_flux_in_frame (combined_footprint_hash, frame_id, star_gaia_id, flux, flux_uncertainty, chi2, relative_loss_differential)
        VALUES (?, ?, ?, ?, ?, ?, ?)
        ON CONFLICT(combined_footprint_hash, frame_id, star_gaia_id) DO UPDATE SET
        flux=excluded.flux, flux_uncertainty=excluded.flux_uncertainty
        """

        cursor.executemany(insert_query, flux_data)

        conn.commit()


def do_star_photometry():
    """
    Here we do a joint deconvolution of all the frames at once, for each star.
    This is equivalent to PSF photometry, but we do it in the same way
    as the final deconvolution to eliminate potential systematics.
    Returns:

    """
    logger = logging.getLogger('lightcurver.star_photometry')
    user_config = get_user_config()
    # so, we need the footprint we are working with.
    # below we will query only the frames we need for each frame, but first we need to
    # query all the frames that we actually started with when defining the footprints:
    # so, let's query those frames, then calculate the footprint.
    frames_ini = get_pandas(columns=['id', 'image_relpath', 'exptime', 'mjd', 'seeing_pixels', 'pixel_scale'],
                            conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_ini['id'].to_list())
    # now we can select the stars we need to do photometry of, within this footprint.
    stars = select_stars(stars_to_use=user_config['stars_to_use_norm'], combined_footprint_hash=combined_footprint_hash)
    logger.info(f"Will do PSF photometry for {len(stars)} stars.")
    # if not re-do ...select only the new frames that do not have a flux measurement yet.
    only_fluxless_frames = not user_config['redo_star_photometry']

    time_now = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")  # for plots
    for i, star in stars.iterrows():
        t0 = time()
        psf_fit_chi2_min, psf_fit_chi2_max = get_chi2_bounds(psf_or_fluxes='psf')
        frames = get_frames_for_star(gaia_id=star['gaia_id'],
                                     psf_fit_chi2_min=psf_fit_chi2_min,
                                     psf_fit_chi2_max=psf_fit_chi2_max,
                                     only_fluxless_frames=only_fluxless_frames,
                                     combined_footprint_hash=combined_footprint_hash)

        if len(frames) == 0:
            # we up to date, nothing to do
            logger.info(f"Star {star['name']}: no new frames to process for this one. Skipping")
            continue
        logger.info(f"Star {star['name']}: has {len(frames)} frames that need a measurement.")
        # build the data for deconvolution
        with h5py.File(user_config['regions_path'], 'r') as h5f:
            data = []
            noisemap = []
            mask = []
            psf = []
            for j, frame in frames.iterrows():
                data.append(h5f[f"{frame['image_relpath']}/data/{star['gaia_id']}"][...])
                noisemap.append(h5f[f"{frame['image_relpath']}/noisemap/{star['gaia_id']}"][...])
                # more difficult for the psf.
                # we need to reconstruct which stars were used in the psf given our
                # accepted stars in 'stars_to_use_psf'.
                stars_psf = select_stars_for_a_frame(frame_id=frame['id'],
                                                     stars_to_use=user_config['stars_to_use_psf'],
                                                     combined_footprint_hash=combined_footprint_hash)
                psf_ref = 'psf_' + ''.join(sorted(stars_psf['name']))
                mask.append(h5f[f"{frame['image_relpath']}/cosmicsmask/{star['gaia_id']}"][...])
                psf.append(h5f[f"{frame['image_relpath']}/{psf_ref}/narrow_psf"][...])
            data, noisemap, mask, psf = np.array(data), np.array(noisemap), np.array(mask), np.array(psf)
            # just like for the PSF, we need to remove the NaNs ...
            isnan = np.where(np.isnan(data) * np.isnan(noisemap))
            data[isnan] = 0.
            noisemap[isnan] = 1e7
            # cosmics: masks are 'true' where cosmic, and we typically want it to "true" for good pixels
            mask = ~(np.array(mask).astype(bool))  # so we invert it.
            # oh, we invert it again to boost the noisemap where mask is False,
            # but better have the masks loaded the right way for the future.
            noisemap[np.where(~mask)[0]] *= 1000.
            # ok now that everything is ready let's get out of the context manager, also to close the file.

        # ready to "deconvolve" using starred!
        result = do_one_deconvolution(data=data, noisemap=noisemap, psf=psf,
                                      subsampling_factor=user_config['subsampling_factor'],
                                      n_iter=user_config['star_deconv_n_iter'])
        # ok, plot the diagnostic
        plot_deconv_dir = user_config['plots_dir'] / 'deconvolutions' / str(combined_footprint_hash)
        plot_deconv_dir.mkdir(exist_ok=True, parents=True)
        loss_history = result['loss_curve']

        plot_file = plot_deconv_dir / f"{time_now}_joint_deconv_star_{star['name']}.jpg"
        plot_joint_deconv_diagnostic(datas=data, noisemaps=noisemap,
                                     residuals=result['residuals'],
                                     chi2_per_frame=result['chi2_per_frame'], loss_curve=loss_history,
                                     save_path=plot_file)

        # now we can insert our results in our dedicated database table.
        loss_index = int(0.9 * np.array(loss_history).size)
        initial_change = np.nanmax(loss_history[:loss_index]) - np.nanmin(loss_history[:loss_index])
        end_change = np.nanmax(loss_history[loss_index:]) - np.nanmin(loss_history[loss_index:])
        relative_loss_differential = float(end_change / initial_change)
        # flux_data should be a list of tuples, each containing (frame_id, star_gaia_id, flux, flux_uncertainty)
        flux_data = []
        # the order of the frames is the same as before, and the same as that of our arrays and fluxes.
        for j, frame in frames.iterrows():
            gaia_id = star['gaia_id']
            frame_id = frame['id']
            flux = float(result['fluxes'][j])
            flux_uncertainty = float(result['fluxes_uncertainties'][j])
            chi2 = float(result['chi2_per_frame'][j])
            flux_data.append((combined_footprint_hash, frame_id, gaia_id, flux, flux_uncertainty,
                              chi2, relative_loss_differential))

        # big insert while updating if value already in DB...
        update_star_fluxes(flux_data)
        # done!
        time_taken = time() - t0
        logger.info(
            f"Measured star {star['name']} in {len(frames)} frames. "
            f"The global reduced chi2 is {result['chi2']:.02f}. "
            f"It took us {time_taken:.01f} seconds to prepare and model the pixels."
        )




import yaml
import os
from pathlib import Path
from astropy.coordinates import SkyCoord
from astropy import units as u

from .exceptions import NoConfigFilePathInEnvironment


def get_user_config():
    """
    This reads the yaml file containing the user config.
    Then processes some of its parameters and builds others.

    Returns:
        dictionary of parameters.
    """
    if 'LIGHTCURVER_CONFIG' not in os.environ:
        raise NoConfigFilePathInEnvironment
    config_path = os.environ['LIGHTCURVER_CONFIG']
    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)

    roi_keys = list(config['ROI'].keys())
    config['roi_name'] = roi_name = roi_keys[0]
    ra, dec = config['ROI'][config['roi_name']]['coordinates']
    config['ROI_ra_deg'] = ra
    config['ROI_dec_deg'] = dec
    config['ROI_SkyCoord'] = SkyCoord(ra*u.deg, dec*u.deg)

    assert 'raw_dirs' in config
    raw_dirs = config['raw_dirs']
    if type(raw_dirs) is list:
        raw_dirs = [Path(pp) for pp in raw_dirs]
    elif type(raw_dirs) is str:
        raw_dirs = [Path(raw_dirs)]
    config['raw_dirs'] = raw_dirs

    assert 'workdir' in config
    config['workdir'] = Path(config['workdir'])
    config['database_path'] = config['workdir'] / 'database.sqlite3'
    config['plots_dir'] = config['workdir'] / 'plots'
    config['logs_dir'] = config['workdir'] / 'logs'
    config['frames_dir'] = config['workdir'] / 'frames'
    config['regions_path'] = config['workdir'] / 'regions.h5'
    config['psfs_path'] = config['workdir'] / 'psfs.h5'
    for directory in [config['plots_dir'], config['logs_dir'], config['frames_dir']]:
        directory.mkdir(parents=True, exist_ok=True)

    # star names: make it a list if user defined a string.
    # e.g. stars_to_use = 'abcd' --> ['a', 'b', 'c', 'd']
    if type(config['stars_to_use_psf']) is str:
        config['stars_to_use_psf'] = [c for c in config['stars_to_use_psf']]
    if type(config['stars_to_use_norm']) is str:
        config['stars_to_use_norm'] = [c for c in config['stars_to_use_norm']]

    return config



import matplotlib.pyplot as plt
from shapely.geometry import Polygon


def plot_footprints(footprint_arrays, common_footprint=None, largest_footprint=None, save_path=None):
    """
    shows the polygons representing the footprints of frames on a plot.

    """
    fig = plt.figure(figsize=(10, 8))
    if (largest_footprint is not None) and (not largest_footprint.is_empty):
        plt.fill(*largest_footprint.exterior.xy, alpha=0.2, color='purple', label='Largest Footprint')
    # Plot each footprint array
    for footprint_array in footprint_arrays:
        polygon = Polygon(footprint_array)
        x, y = polygon.exterior.xy
        plt.plot(x, y, color='gray', lw=0.5)
    if (common_footprint is not None) and (not common_footprint.is_empty):
        plt.fill(*common_footprint.exterior.xy, alpha=0.5, color='blue', label='Common Footprint')
    plt.legend()
    plt.xlabel('R.A.')
    plt.ylabel('Dec.')
    if save_path is not None:
        plt.tight_layout()
        plt.savefig(save_path, bbox_inches='tight', pad_inches=0.)
        plt.close()
    else:
        return fig, plt.gca()





import h5py
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
import astropy.units as u
from astropy.coordinates import SkyCoord
from astroscrappy import detect_cosmics
from astropy.time import Time
import logging
# to suppress the warnings telling us we don't have the parallax:
# we do not care about the radial component of proper motion, we just want
# roughly centered stars.
import warnings
import erfa

from ..structure.user_config import get_user_config
from ..structure.database import get_pandas, query_stars_for_frame_and_footprint
from ..utilities.footprint import get_combined_footprint_hash


def extract_stamp(data, header, exptime, sky_coord, cutout_size, background_rms_electron_per_second):
    """
    :param data: 2d numpy array containing the full image
    :param header: fits header for WCS info
    :param exptime: float, exposure time to convert from e-/s to e- and back.
    :param sky_coord: astropy SkyCoord: center of cutout
    :param cutout_size: int, pixels
    :return: 2d cutout array, 2d cutout noisemap array, wcs string of cutout
    """

    wcs = WCS(header)
    data_cutout = Cutout2D(data, sky_coord, cutout_size, wcs=wcs, mode='partial')
    # let's also carry the WCS of the cutouts
    wcs_header = data_cutout.wcs.to_header()
    wcs_header_string = wcs_header.tostring()

    # now just take the numpy array
    data_cutout_electrons = exptime * data_cutout.data

    noisemap_electrons = ((exptime * background_rms_electron_per_second)**2 + np.abs(data_cutout_electrons))**0.5
    # remove zeros if there are any ...
    noisemap_electrons[noisemap_electrons < 1e-7] = 1e-7

    return data_cutout.data, noisemap_electrons / exptime, wcs_header_string


def extract_all_stamps():
    """
    This is the routine that the workflow manager will call.
    It interfaces with the user config and the database to locate the
    files and objects to extract, then extracts to a hdf5 file.

    Returns:
        Nothing
    """
    logger = logging.getLogger('lightcurver.cutout_making')
    user_config = get_user_config()

    # where we'll save our stamps
    regions_file = user_config['regions_path']

    # query frames
    frames_to_process = get_pandas(columns=['id', 'image_relpath', 'exptime', 'mjd',
                                            'background_rms_electron_per_second'],
                                   conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])

    # we'll need to know what combined_footprint we're working with.
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_to_process['id'].to_list())
    logger.info(
        f'Will extract cutouts from (potentially) {len(frames_to_process)} frames, in file {regions_file}. '
        f'The combined footprint hash is {combined_footprint_hash}.'
    )
    # suppress the no parallax warning in proper motion corrections:
    warnings.filterwarnings('ignore', category=erfa.ErfaWarning)

    with h5py.File(regions_file, 'a') as reg_f:
        for i, frame in frames_to_process.iterrows():
            # check what stars need be extracted
            stars = query_stars_for_frame_and_footprint(frame_id=frame['id'],
                                                        combined_footprint_hash=combined_footprint_hash)

            # chance to skip this frame if both 'not redo' and 'all was extracted'
            if (not user_config['redo_stamp_extraction']) and (frame['image_relpath'] in reg_f.keys()):
                if not len(reg_f[frame['image_relpath']]['data'].keys()) == len(stars) + 1:
                    # not the right number of cutouts ...have to extract.
                    pass
                else:
                    # right number of cutouts, but gotta make sure the keys match.
                    keys = reg_f[frame['image_relpath']]['data'].keys()
                    all_there = True
                    for j, star in stars.iterrows():
                        if star['gaia_id'] not in keys:
                            all_there = False
                            break
                    if 'ROI' not in keys:
                        all_there = False
                    if all_there:
                        # all good, skip to next frame.
                        logger.info(f"Frame with id {frame['id']} likely already extracted, skipping.")
                        continue
            image_file = user_config['workdir'] / frame['image_relpath']
            data, header = fits.getdata(image_file), fits.getheader(image_file)
            global_rms = frame['background_rms_electron_per_second']
            # organize hdf5 file
            if frame['image_relpath'] not in reg_f.keys():
                frame_set = reg_f.create_group(frame['image_relpath'])
            else:
                frame_set = reg_f[frame['image_relpath']]
            if 'data' not in frame_set.keys():
                data_set = frame_set.create_group('data')
            else:
                data_set = frame_set['data']
            if 'noisemap' not in frame_set.keys():
                noise_set = frame_set.create_group('noisemap')
            else:
                noise_set = frame_set['noisemap']
            if 'wcs' not in frame_set.keys():
                wcs_set = frame_set.create_group('wcs')
            else:
                wcs_set = frame_set['wcs']
            if 'cosmicsmask' not in frame_set.keys():
                cosmic_mask = frame_set.create_group('cosmicsmask')
            else:
                cosmic_mask = frame_set['cosmicsmask']

            if user_config['redo_stamp_extraction'] or ('ROI' not in cosmic_mask.keys()):
                # extract the ROI -- assuming no proper motion.
                cutout, noisemap, wcs_str = extract_stamp(data=data, header=header,
                                                          exptime=frame['exptime'],
                                                          sky_coord=user_config['ROI_SkyCoord'],
                                                          cutout_size=user_config['stamp_size_ROI'],
                                                          background_rms_electron_per_second=global_rms)
                # clean the cosmics
                if user_config['clean_cosmics']:
                    mask, cleaned = detect_cosmics(cutout, invar=noisemap**2)
                else:
                    mask = np.zeros_like(cutout, dtype=bool)
                    cleaned = cutout
                if 'ROI' in data_set:
                    del data_set['ROI']
                data_set['ROI'] = cleaned
                if 'ROI' in noise_set:
                    del noise_set['ROI']
                noise_set['ROI'] = noisemap
                if 'ROI' in wcs_set:
                    del wcs_set['ROI']
                wcs_set['ROI'] = wcs_str
                if 'ROI' in cosmic_mask:
                    del cosmic_mask['ROI']
                cosmic_mask['ROI'] = mask

            # set proper motion to 0 when not available
            if len(stars) > 0:  # if 0 stars, then frame will not be queried downstream.
                stars.loc[np.isnan(stars['pmra']), 'pmra'] = 0.0
                stars.loc[np.isnan(stars['pmdec']), 'pmdec'] = 0.0
            elif len(stars) == 0:
                logger.warning(
                    f"Frame with id {frame['id']} has no star available for extraction. "
                    "This is fine, it will simply not be used in the downstream steps."
                )
            # extract the stars
            for j, star in stars.iterrows():
                star_name = str(star['gaia_id'])
                if user_config['redo_stamp_extraction'] or (star_name not in cosmic_mask.keys()):
                    # make a sky coord with our star and its proper motion
                    star_coord = SkyCoord(ra=star['ra'] * u.deg,
                                          dec=star['dec'] * u.deg,
                                          pm_ra_cosdec=star['pmra'] * u.mas / u.yr,
                                          pm_dec=star['pmdec'] * u.mas / u.yr,
                                          frame='icrs',
                                          obstime=Time(star['ref_epoch'], format='decimalyear'))
                    # then correct the proper motion
                    obs_epoch = Time(frame['mjd'], format='mjd')
                    corrected_coord = star_coord.apply_space_motion(new_obstime=obs_epoch)
                    cutout, noisemap, wcs_str = extract_stamp(data=data, header=header,
                                                              exptime=frame['exptime'],
                                                              sky_coord=corrected_coord,
                                                              cutout_size=user_config['stamp_size_stars'],
                                                              background_rms_electron_per_second=global_rms)

                    # again, clean the cosmics.
                    if user_config['clean_cosmics']:
                        mask, cleaned = detect_cosmics(cutout, invar=noisemap**2)
                    else:
                        mask = np.zeros_like(cutout, dtype=bool)
                        cleaned = cutout
                    if star_name in data_set:
                        del data_set[star_name]
                    data_set[star_name] = cleaned
                    if star_name in noise_set:
                        del noise_set[star_name]
                    noise_set[star_name] = noisemap
                    if star_name in wcs_set:
                        del wcs_set[star_name]
                    wcs_set[star_name] = wcs_str
                    if star_name in cosmic_mask:
                        del cosmic_mask[star_name]
                    cosmic_mask[star_name] = mask

            logger.info(f"Frame with id {frame['id']}: completed making of cutouts.")



import h5py
import numpy as np
import logging

from ..structure.database import get_pandas, execute_sqlite_query
from ..utilities.footprint import get_combined_footprint_hash
from ..utilities.chi2_selector import get_chi2_bounds
from ..structure.user_config import get_user_config


def get_frames_for_roi(combined_footprint_hash, psf_fit_chi2_min, psf_fit_chi2_max, **constraints_on_frame_columns_dict):
    """
    Retrieves frames and associated PSFs (built with stars of the given footprint)
    provided that those frames have a PSF with a chi2 between psf_fit_chi2_min and psf_fit_chi2_max.
    Optionally, can filter to include only frames without a flux measurement.

    :param combined_footprint_hash: int, the hash of the combined footprint we are processing.
    :param psf_fit_chi2_min: The minimum acceptable chi2 value for the PSF fit.
    :param psf_fit_chi2_max: The maximum acceptable chi2 value for the PSF fit.
    :param constraints_on_frame_columns_dict: a dictionary, with keys identical to that of the frames table and values
                                              intervals of allowed values. E.g., {'seeing_arcseconds': (0, 1.1), ...}
    :return: A pandas dataframe of frames and associated PSFs that meet the criteria.
    """
    query = """
    SELECT f.*, ps.*, nc.*
    FROM frames f
    JOIN PSFs ps ON f.id = ps.frame_id 
    JOIN normalization_coefficients nc ON f.id = nc.frame_id AND nc.combined_footprint_hash = ps.combined_footprint_hash
    WHERE nc.combined_footprint_hash = ?
    AND ps.chi2 BETWEEN ? AND ?
    """

    # params of the query:
    params = [combined_footprint_hash, psf_fit_chi2_min, psf_fit_chi2_max]

    # append constraints based on the provided constraints
    for column, (min_val, max_val) in constraints_on_frame_columns_dict.items():
        query += f" AND f.{column} BETWEEN ? AND ?"
        params.extend([min_val, max_val])

    return execute_sqlite_query(query, tuple(params), is_select=True, use_pandas=True)


def fetch_and_adjust_zeropoints(combined_footprint_hash):
    """
    Just a helper function.
    We query our normalization coefficients and zeropoints.
    We adjust the zeropoints by the normalization coefficient.
    We check that the scatter in zeropoints after normalizing is smaller than the scatter before.
    (just a sanity check really)

    we return the global zeropoint of normalized data, and a scatter we can use as STATISTICAL uncertainty.

    Params:
        combined_footprint_hash: as usual the hash of the footprint we're working with at the moment.

    """

    zeropoint_query = """
    SELECT 
        az.frame_id, 
        az.zeropoint, 
        az.zeropoint_uncertainty, 
        nc.coefficient
    FROM 
        approximate_zeropoints az
    JOIN 
        normalization_coefficients nc ON az.frame_id = nc.frame_id
    AND
        az.combined_footprint_hash = nc.combined_footprint_hash
    WHERE 
        az.combined_footprint_hash = ? 
    """
    zeropoints_data = execute_sqlite_query(zeropoint_query, (combined_footprint_hash,), is_select=True, use_pandas=True)

    if zeropoints_data.empty:
        return None, None

    # convert normalization coefficient to magnitude adjustment,
    # then adjust the zeropoint by subtracting the magnitude adjustment
    zeropoints_data['norm_adjustment'] = -2.5 * np.log10(zeropoints_data['coefficient'])
    zeropoints_data['adjusted_zeropoint'] = zeropoints_data['zeropoint'] + zeropoints_data['norm_adjustment']

    zp_scatter_not_normalized = zeropoints_data['zeropoint'].std()
    zp_scatter_normalized = zeropoints_data['adjusted_zeropoint'].std()
    global_zp = zeropoints_data['adjusted_zeropoint'].median()

    message = "The scatter in zeropoints before normalizing is lower than after normalizing? Not normal, investigate."
    assert zp_scatter_normalized < zp_scatter_not_normalized, message

    return global_zp, zp_scatter_normalized


def prepare_roi_deconv_file():
    """
    This is called by the workflow manager.
    Given the previous steps, selects the frames that have the necessary quantities calculated
    (PSF, normalization coefficient).
    Then makes a new hdf5 file with the cutouts of the ROI (both data and noisemaps,
    divided by normalization coefficients), and the PSFs.
    Returns:
        Nothing
    """
    logger = logging.getLogger('lightcurver.roi_deconv_file_preparation')
    user_config = get_user_config()

    frames_ini = get_pandas(columns=['id'],
                            conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_ini['id'].to_list())

    psf_fit_chi2_min, psf_fit_chi2_max = get_chi2_bounds(psf_or_fluxes='psf')

    roi_constraints = user_config['constraints_on_frame_columns_for_roi']
    frames = get_frames_for_roi(combined_footprint_hash=combined_footprint_hash,
                                psf_fit_chi2_min=psf_fit_chi2_min,
                                psf_fit_chi2_max=psf_fit_chi2_max,
                                **roi_constraints)
    # ok, this frames database has everything: the PSF to use, the norm coeff, etc.
    logger.info(f'Preparing calibrated cutouts of the ROI from {len(frames)} frames.')
    # so, just like when we did photometry of stars, build the data for deconvolution
    with h5py.File(user_config['regions_path'], 'r') as h5f:
        data = []
        noisemap = []
        mask = []
        psf = []
        frame_id = []
        subsampling_factors = []  # should be a unique value but this way we can check
        angles_to_north = []
        # we'll include some more stuff for reference
        seeing, pixel_scale, wcs, mjd, exptime, sky_level_electron_per_second = [], [], [], [], [], []
        normalization_uncertainty = []
        for j, frame in frames.iterrows():
            coefficient = frame['coefficient']
            data.append(h5f[f"{frame['image_relpath']}/data/ROI"][...] / coefficient)
            noisemap.append(h5f[f"{frame['image_relpath']}/noisemap/ROI"][...] / coefficient)
            mask.append(h5f[f"{frame['image_relpath']}/cosmicsmask/ROI"][...])
            psf_ref = frame['psf_ref']
            psf.append(h5f[f"{frame['image_relpath']}/{psf_ref}/narrow_psf"][...])
            subsampling_factors.append(h5f[f"{frame['image_relpath']}/{psf_ref}/subsampling_factor"][...])
            seeing.append(frame['seeing_arcseconds'])
            pixel_scale.append(frame['pixel_scale'])
            wcs.append(h5f[f"{frame['image_relpath']}/wcs/ROI"][()])
            exptime.append(frame['exptime'])
            sky_level_electron_per_second.append(frame['sky_level_electron_per_second'])
            mjd.append(frame['mjd'])
            frame_id.append(frame['id'])
            normalization_uncertainty.append(frame['coefficient_uncertainty'])
            angles_to_north.append(frame['angle_to_north'])
        data, noisemap, mask, psf = np.array(data), np.array(noisemap), np.array(mask), np.array(psf)
        # just like for the PSF, we need to remove the NaNs ...
        isnan = np.where(np.isnan(data) * np.isnan(noisemap))
        data[isnan] = 0.
        noisemap[isnan] = 1e7
        # cosmics: masks are 'true' where cosmic, and we typically want it to "true" for good pixels
        mask = ~(np.array(mask).astype(bool))  # so we invert it.
        # oh, we invert it again to boost the noisemap where mask is False,
        # but better have the masks loaded the right way for the future.
        noisemap[~mask] *= 1000.
        # ok now that everything is ready let's get out of the context manager, also to close the file,
        # and we can open the deconvolution ready file.

    # get the zeropoint:
    global_zp, global_zp_scatter = fetch_and_adjust_zeropoints(combined_footprint_hash=combined_footprint_hash)

    # where we save the ready to deconvolve cutouts:
    save_path = user_config['prepared_roi_cutouts_path']
    
    if save_path is None:
        roi = user_config['roi_name']
        save_path = user_config['workdir'] / 'prepared_roi_cutouts' / f"cutouts_{combined_footprint_hash}_{roi}.h5"
    save_path.parent.mkdir(exist_ok=True, parents=True)
    with h5py.File(save_path, 'w') as f:
        f['frame_id'] = np.array(frame_id)
        f['data'] = np.array(data)
        f['noisemap'] = np.array(noisemap)
        f['psf'] = np.array(psf)
        f['seeing'] = np.array(seeing)
        f['sky_level_electron_per_second'] = np.array(sky_level_electron_per_second)
        f['mjd'] = np.array(mjd)
        f['global_zeropoint'] = np.array(float(global_zp))
        f['global_zeropoint_scatter'] = np.array(float(global_zp_scatter))
        f['relative_normalization_error'] = np.array(normalization_uncertainty)
        f['wcs'] = np.array(wcs)
        f['pixel_scale'] = np.array(pixel_scale)
        f['subsampling_factor'] = np.array(subsampling_factors)
        f['angle_to_north'] = np.array(angles_to_north)

    logger.info(f'Wrote the h5 file containing the calibrated cutouts at {save_path}.')



import matplotlib.pyplot as plt
import numpy as np


def plot_psf_diagnostic(datas, noisemaps, residuals, full_psf,
                        loss_curve=None, masks=None, names=None,
                        diagnostic_text=None,
                        save_path=None):
    """
    a utility to plot a summary of a psf modelling: plots the stars, noisemaps, and residuals.
    optionally, one can pass the loss curve (to make sure it looks like the model converged),
    names (one name per star), masks (for visualization, will appear as holes in the noisemaps),
    and some diagnostic text to be displayed at the top right.
    Args:
        datas: 3d array of shape (N, nx, ny): N stars, one slice per star.
        noisemaps: same as above but for the noisemaps
        residuals: same as above but for the residuals
        full_psf: full pixelated psf model
        loss_curve: optional, 1D array containing the evolution of the loss during optimization
        masks: optional, 3d array of shape (N, nx, ny): the masks that were used during optimization.
        names: optional, list of strings: for identifying stars.
        diagnostic_text: optional, string, with line breaks (max ~ 20 chars per line)
        save_path: optional, string or path, where to save the plot.

    Returns: None

    """
    plt.style.use('dark_background')
    cmap = 'viridis'
    cmap_residuals = 'coolwarm'
    text_color = 'white'
    text_size = 11
    single_letter_text_size = 14
    info_box_text_size = 6

    N = len(datas)
    if names is not None:
        assert N == len(names)

    sub_size = 3
    fig, ax = plt.subplots(3, N+1, figsize=((N+1)*sub_size, 3*sub_size))

    for i in range(N):
        for j in range(3):
            ax[j, i].axis('off')
            ax[j, i].set_aspect('equal')
            if j == 0:
                ax[j, i].imshow(datas[i], cmap=cmap)
                if names is not None:
                    ax[j, i].text(0.5, 0.02, names[i],
                                  horizontalalignment='center',
                                  verticalalignment='bottom',
                                  transform=ax[j, i].transAxes,
                                  color=text_color, fontsize=single_letter_text_size,
                                  weight='bold')
            elif j == 1:
                ax[j, i].imshow(noisemaps[i], cmap=cmap)
                ax[j, i].text(0.5, 0.02, 'noisemap, mask',
                              horizontalalignment='center',
                              verticalalignment='bottom',
                              transform=ax[j, i].transAxes,
                              color=text_color, fontsize=text_size,
                              weight='bold')
            elif j == 2:
                res = np.array(residuals[i])  # explicit casting, jax stuff
                if masks is not None:
                    mask = np.array(masks[i]).astype(bool)
                    res[np.where(~mask)] = np.nan
                ax[j, i].imshow(res, cmap=cmap_residuals)
                ax[j, i].text(0.5, 0.02, 'residuals',
                              horizontalalignment='center',
                              verticalalignment='bottom',
                              transform=ax[j, i].transAxes,
                              color=text_color, fontsize=text_size,
                              weight='bold')

    # info box
    ax[0, N].text(0.1, 0.99, diagnostic_text,
                  horizontalalignment='left',
                  verticalalignment='top',
                  fontsize=info_box_text_size,
                  color='white')
    ax[0, N].axis('off')

    # loss plot
    if loss_curve is not None:
        ax[1, N].plot(loss_curve, color='white')
        ax[1, N].text(0.5, 0.99, 'loss',
                      horizontalalignment='center',
                      verticalalignment='top',
                      transform=ax[1, N].transAxes,
                      color='white', fontsize=text_size,
                      weight='bold')

        ax[1, N].axis('off')
    # psf model plot
    ax[2, N].imshow(full_psf, cmap=cmap, aspect='auto')
    ax[2, N].axis('off')
    ax[2, N].text(0.5, 0.01, 'Full PSF',
                  horizontalalignment='center',
                  verticalalignment='bottom',
                  transform=ax[2, N].transAxes,
                  color=text_color, fontsize=text_size,
                  weight='bold')
    ax[2, N].set_aspect('equal')

    plt.subplots_adjust(wspace=0, hspace=0)
    if save_path is not None:
        plt.savefig(save_path, pad_inches=0, bbox_inches='tight')
        plt.close()
    else:
        return fig, ax
